(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

module type SharedLocalType = sig
  type shared
  type component
  val compare_shared: shared->shared->int 
  val compare_component: component->component->int
  val string_of_shared: shared->string
  val string_of_component: component->string
end

module type S=sig
  type shared
  type component
  type setOfComponent
  type setOfShared
  type state
  type setOfStates
  val empty_setOfStates: setOfStates
  val is_empty_setOfStates: setOfStates->bool
  val single_state: shared -> component array -> state
  val singleton_setOfStates: shared -> component array -> setOfStates
  val choose_state_from_setOfStates: setOfStates -> state
  val state_dimension: state->int
  val add_state_setOfStates: state->setOfStates->setOfStates
  val union_setOfStates_setOfStates: setOfStates->setOfStates->setOfStates
  val intersection_setOfStates_setOfStates: setOfStates->setOfStates->setOfStates
  val diff_setOfStates_setOfStates: setOfStates->setOfStates->setOfStates
  val string_of_state:state->string
  val string_of_setOfStates: setOfStates->string
  val string_of_trace: setOfStates list -> string

  type transition
  type transitionSet
  val empty_transitionSet : transitionSet
  val is_empty_transitionSet:transitionSet->bool
  val single_transition:shared->component->int->shared->component->transition
  val singleton_transition: shared->component->int->shared->component->transitionSet
  val add_transition_transitionSet:transition->transitionSet->transitionSet
  val union_transitionSet_transitionSet:transitionSet->transitionSet->transitionSet
  val apply_transitionSet: setOfStates -> transitionSet -> setOfStates
  val apply_reversed_transitionSet: setOfStates -> transitionSet -> setOfStates
  val reverse_transitionSet: transitionSet -> transitionSet
  val string_of_transitionSet: transitionSet->string

  type abstractElement
  val least_abstractElement: abstractElement
  val string_of_abstractElement:abstractElement->string
  val abstract_setOfStates: setOfStates -> abstractElement
  val abstract_setOfStates_Minus_setOfStates: setOfStates->setOfStates->(abstractElement*setOfStates)
  val join_abstract_element_cartabs_setOfStates_minus_setOfStates: abstractElement->setOfStates->setOfStates->(abstractElement*setOfStates)
  val le_abstract_elements: abstractElement -> abstractElement -> bool
  val join_abstract_elements:abstractElement->abstractElement->abstractElement
  val concretize_abstractElement_union_setOfStates: abstractElement->setOfStates->setOfStates
  exception ExceptionSetNotFound
  val exception_set: abstractElement-> setOfStates -> setOfStates ->setOfStates
  val no_mutex_setOfStates: setOfComponent->setOfComponent->setOfShared->int->setOfStates
  val no_assertion_setOfStates: setOfComponent->setOfComponent->setOfShared->int->setOfStates
end

module MakeOrderedLocal(SL:SharedLocalType): (CartesianProduct.OrderedTypeWithPrinting with type t=SL.component) = struct
  type t=SL.component
  let compare=SL.compare_component
  let to_string=SL.string_of_component
end


module MakeOrderedShared(SL:SharedLocalType): (Set.OrderedType with type t=SL.shared) = struct
  type t=SL.shared
  let compare=SL.compare_shared
end

module Make (SL:SharedLocalType) : S with type shared=SL.shared and type component=SL.component and type setOfComponent=Set.Make(MakeOrderedLocal(SL)).t and type setOfShared=Set.Make(MakeOrderedShared(SL)).t  = struct
  (* The type of a shared state *)
  type shared = SL.shared

  (* the type of a local state *)
  type component=SL.component

  (* the type of a program state *)
  type state=SL.shared*(SL.component array)

  module OrderedSharedState = struct
    type t=shared
    let compare = SL.compare_shared
  end

  module MapFromSharedStates = ExtMap.Make(OrderedSharedState)
    
  module CartProdOfLocals=CartesianProduct.Make(MakeOrderedLocal(SL))
  type cartProdOfLocals=CartProdOfLocals.t

  module SetOfComponent=Set.Make(MakeOrderedLocal(SL))
  type setOfComponent=SetOfComponent.t

  module SetOfShared=Set.Make(MakeOrderedShared(SL))
  type setOfShared=SetOfShared.t

  (* The type of a set of states *)
  type setOfStates=(CartProdOfLocals.setOfTuples) MapFromSharedStates.t

  (* The empty set of states *)
  let empty_setOfStates=MapFromSharedStates.empty

  (* Check whether a given set of states is empty or not *)
  let is_empty_setOfStates (s:setOfStates)=MapFromSharedStates.for_all s (fun g ls ->(CartProdOfLocals.is_empty_setOfTuples ls))


  (* Returns a string representation of a string. *)
  let string_of_state (g,l)=
    let sg=SL.string_of_shared g in
    let len=Array.length l in
      if len=0 then sg else
	"("^sg^","^
	  (let last=l.(len-1) in
	   let rest=Array.sub l 0 (len-1) in
	     Array.fold_right
	       (fun li sofar -> (SL.string_of_component li)^","^sofar) rest (SL.string_of_component last)
	  )
	^")"

  (* creates a string from a single binding shared |-> set of tuples. *)
  let string_of_shared_locals g sot =
    "{"^(SL.string_of_shared g)^"}x"^
      (CartProdOfLocals.string_of_setOfTuples_possibly_with_brackets sot)

  (* Convert a set of states into a string. *)
  let string_of_setOfStates sos =
    if MapFromSharedStates.is_empty sos then "{}" else
      MapFromSharedStates.fold 
	(fun g sot sofar -> 
	   (string_of_shared_locals g sot)^(if (String.length sofar)>0 then "u"^sofar else "")) sos ""

  (* creates a single state *)
  let single_state (g:shared) (ca:component array) : state = (g,ca)

  (* Returns any arbitrary state from a set of states. Throws Not_found if the set of states is empty. *)
  let choose_state_from_setOfStates (sos:setOfStates) = 
    let (g,sot)=MapFromSharedStates.exists sos (fun g sot -> not (CartProdOfLocals.is_empty_setOfTuples sot))
    in (g,CartProdOfLocals.choose_tuple_from_setOfTuples sot)

  (* Returns the number of locals in a state. *)
  let state_dimension s = Array.length (snd s)

  (* creates a singleton from the shared and the local part *)
  let singleton_setOfStates (g:shared) (ca:component array): setOfStates = 
    MapFromSharedStates.add g (CartProdOfLocals.singleton_setOfTuples ca) empty_setOfStates

  (* Adds an element to the set of states *)
  let add_state_setOfStates (g,l) s = 
    let new_sot=
      try 
	let sot=MapFromSharedStates.find g s 
	in CartProdOfLocals.add_tuple_to_setOfTuples l sot
      with Not_found -> CartProdOfLocals.singleton_setOfTuples l
    in MapFromSharedStates.add g new_sot s

  (* Returns the union of two sets of states. *)
  let rec union_setOfStates_setOfStates s1 s2 =
    try
      let (g,sot2)=MapFromSharedStates.any_binding s2
      in let new_s1=
	  (try
	     let sot1=MapFromSharedStates.find g s1
	     in MapFromSharedStates.add g (CartProdOfLocals.union_setOfTuples_setOfTuples sot1 sot2) s1
	   with Not_found ->MapFromSharedStates.add g sot2 s1)
      in let new_s2=MapFromSharedStates.remove g s2
      in union_setOfStates_setOfStates new_s1 new_s2
    with Not_found -> s1

  (* Returns the intersection of two sets of states *)
  let intersection_setOfStates_setOfStates s1 s2 =
    let f scp1 scp2 = 
      let scp=CartProdOfLocals.inter_setOfTuples_setOfTuples scp1 scp2 in 
	if CartProdOfLocals.is_empty_setOfTuples scp then None else Some scp
    in MapFromSharedStates.intersect_apply s1 s2 f

  (* Returns the difference of two sets of states *)
  let diff_setOfStates_setOfStates s1 s2 =
(*    let _= Debug.print_debug ("AbstractConcreteElement.diff_setOfStates_setOfStates started on "^(string_of_setOfStates s1)^" and "^(string_of_setOfStates s2)^".\n") in let result=*)
    let f scp1 scp2 = 
      let scp=CartProdOfLocals.diff_setOfTuples_setOfTuples scp1 scp2 in 
	if CartProdOfLocals.is_empty_setOfTuples scp then None else Some scp
    in MapFromSharedStates.diff_apply s1 s2 f
(*    in let _= Debug.print_debug ("AbstractConcreteElement.diff_setOfStates_setOfStates finished with "^(string_of_setOfStates result)^".\n") in result *)

  (* The type of a single transition. *)
  type transition=shared*component*int*shared*component
    (** Type of a single transition. (g,l,i,g',l') means a transition from 
	a state with shared part g and local part l of the ith process to the
	state where the shared part is changed to g' and the local part of the
	ith process is changed to l'. *)
    
  (* For the sorting purposes inside a set we split the transition into the 
     head and the remainder. The head contains the thread number and the 
     source shared state. The remainder contains the source local state, 
     the target shared state, the target local state.
  *)

  module OrderedTransitionHead = 
    struct
      type t= shared
      let compare = SL.compare_shared
    end

  module OrderedTransitionRemainder =
    struct
      type t=component*int*shared*component
      (* source local state, thread, target shared state, target local state*)
      let compare (l1,i1,g1',l1') (l2, i2,g2',l2') =
	if i1<i2 then -1 else if i1>i2 then 1 else
	let cmp_g'=SL.compare_shared g1' g2' in
	if cmp_g'<0 then -1 else if cmp_g'>0 then 1 else
	let cmp_l=SL.compare_component l1 l2 in
	if cmp_l<0 then -1 else if cmp_l>0 then 1 else
	let cmp_l'=SL.compare_component l1' l2' in
	if cmp_l'<0 then -1 else if cmp_l'>0 then 1 else 0
    end

  module TransitionRemainderSet = ExtSet.Make(OrderedTransitionRemainder)
  type transitionRemainderSet = TransitionRemainderSet.t
  module TransitionHeadMap=ExtMap.Make(OrderedTransitionHead)
  type transitionSet=transitionRemainderSet TransitionHeadMap.t
    (* Type of a set of transitions *)

  let empty_transitionSet = TransitionHeadMap.empty

  let is_empty_transitionSet = TransitionHeadMap.is_empty


  (* Convert a transition into a string. *)
  let string_of_transition (g,l,i,g',l') =
    "("^(SL.string_of_shared g)^","^(SL.string_of_component l)^","^(string_of_int i)^","^(SL.string_of_shared g')^","^(SL.string_of_component l')^")"

  (* Convert a set of transitions remainders into a comma-separated string 
     given a fixed source shared state and a fixed thread number. 
     Expect a nonempty set.
   *)
  let string_of_transitionRemainderSet g trs =
    let ((l,i,g',l'),rest)=TransitionRemainderSet.remove_and_return_min_elt trs
    in TransitionRemainderSet.fold (fun (l,i,g',l') sofar -> sofar^","^(string_of_transition (g,l,i,g',l'))) rest (string_of_transition (g,l,i,g',l'))

(*
    let f (l,i,g',l') sofar = 
      let new_str=(string_of_transition (g,l,i,g',l')) in
      if ((String.length sofar)=0) then new_str else (sofar^","^new_str)
    in TransitionRemainderSet.fold f trs ""
*)

  (* Convert a set of transitions into a string *)
  let string_of_transitionSet ts =
    let f g trs sofar = 
      let new_str=string_of_transitionRemainderSet g trs in
	if ((String.length sofar)=0) then new_str else (sofar^","^new_str)
    in ("{"^(TransitionHeadMap.fold f ts "")^"}")
      
  (* Creating a single transition. *)
  let single_transition g l i g' l' = (g,l,i,g',l')

  (* return a transition set consisting of one single transition *)
  let singleton_transition g l i g' l' : transitionSet =
    let srs = TransitionRemainderSet.singleton (l,i,g',l') in
    TransitionHeadMap.add g srs TransitionHeadMap.empty
  
  (* add a transition to a transition set *)
  let add_transition_transitionSet ((g,l,i,g',l'):transition) ts =
(*
    Debug.print_debug ("AbstractConcreteElement::add_transition_transitionSet started on a Head Set of cardinality "^(string_of_int (TransitionHeadMap.cardinal ts))^"\n");
    let result=
*)
      let tr_new_rem=
	try
	  TransitionRemainderSet.add (l,i,g',l') (TransitionHeadMap.find g ts)
	with Not_found -> TransitionRemainderSet.singleton (l,i,g',l')
      in TransitionHeadMap.add g tr_new_rem ts
(*
    in Debug.print_debug ("AbstractConcreteElement::add_transition_transitionSet finished on a Head Set of cardinality "^(string_of_int (TransitionHeadMap.cardinal result))^"\n"); result
*)

  (* Take the union of two transition sets. *)
  let union_transitionSet_transitionSet = 
    TransitionHeadMap.union TransitionRemainderSet.union

  (* apply_transitionRemainderSet_setOfTuples trs sot returns a set of states
     from taking all transitions remainders (l,i,g',l'), applying the change
     (l,i,l') to sot, attaching the result to g' and taking the union of the 
     sets of states.
  *)
  let apply_transitionRemainderSet_setOfTuples trs sot =
    TransitionRemainderSet.fold 
      (fun (l,i,g',l') sofar ->
	 let sot'=CartProdOfLocals.apply_transition_setOfTuples sot i l l' in
	   if CartProdOfLocals.is_empty_setOfTuples sot' then sofar else
	     let sos'=MapFromSharedStates.add g' sot' empty_setOfStates in
	       union_setOfStates_setOfStates sos' sofar
      ) trs empty_setOfStates

  (* apply_transitionSet sos ts returns a set of states which results from 
     applying all transitions from ts to sos and uniting the results. *)
  let apply_transitionSet sos ts =
(*
    let _= Debug.print_debug ("apply_transitionSet started with sos="^(string_of_setOfStates sos)^" and ts="^(string_of_transitionSet ts)^".\n") in
    let result =
*)
    TransitionHeadMap.fold
      (fun g trs sofar ->
	 try 
	   let sot= MapFromSharedStates.find g sos in
	   let sos'=apply_transitionRemainderSet_setOfTuples trs sot in
	     union_setOfStates_setOfStates sos' sofar
	 with Not_found -> sofar
      ) ts empty_setOfStates
(*    in
    let _= Debug.print_debug ("apply_transitionSet returns "^(string_of_setOfStates result)^".\n") in
      result
*)

  (* apply_reversed_transitionRemainderSet sos' trs 
     returns a set of tuples that result from taking any state (g,l1,...,ln)
     from sos', taking any transition remainder (l,i,g',l') and producing
        (l1,...,l(i-1),l,l(i+1),...,ln) if li = l' and g'=g and
        nothing otherwise
     and taking the union.
  *)
  let apply_reversed_transitionRemainderSet sos' trs =
    TransitionRemainderSet.fold 
      (fun (l,i,g',l') sofar -> 
	 try
	   let sot'=MapFromSharedStates.find g' sos' in
	   let sot =CartProdOfLocals.apply_transition_setOfTuples sot' i l' l in
	     CartProdOfLocals.union_setOfTuples_setOfTuples sot sofar
	 with Not_found -> sofar
      ) trs CartProdOfLocals.empty_setOfTuples

  (* apply_reversed_transitionSet sos ts returns a set of states which 
     results from applying the reverse of all transitions from ts to sos and 
     uniting the results. *)
  let apply_reversed_transitionSet sos' ts =
    TransitionHeadMap.fold 
      (fun g trs sofar ->
	 let sot'=apply_reversed_transitionRemainderSet sos' trs in
	  if CartProdOfLocals.is_empty_setOfTuples sot' then sofar
	  else MapFromSharedStates.add g sot' sofar
      ) ts empty_setOfStates

  (* reverse_transitionSet ts 
     transforms each transition (g,l,i,g',l') into (g',l',i,g,l)
  *)
  let reverse_transitionSet ts =
    TransitionHeadMap.fold
      (fun g trs sofar ->
	 TransitionRemainderSet.fold
	   (fun (l,i,g',l') sofar_r->
	      add_transition_transitionSet (g',l',i,g,l) sofar_r
	   ) trs sofar
      ) ts empty_transitionSet


  (* Type of the elements of the abstract domain. *)
  type abstractElement = (CartProdOfLocals.t) MapFromSharedStates.t

  (* The bottom of the lattice. *)
  let least_abstractElement=MapFromSharedStates.empty

  (* Convert an abstract element into a string. *)
  let string_of_abstractElement (ae:abstractElement):string = 
    if MapFromSharedStates.is_empty ae then "{}" else 
      MapFromSharedStates.fold (fun g cp sofar -> "{"^(SL.string_of_shared g)^"}x"^(CartProdOfLocals.string_of_CP cp)^(if (String.length sofar)>0 then "u"^sofar else "")) ae "" 

  (* Compute the abstraction of a set of states. *)
  let abstract_setOfStates = MapFromSharedStates.map CartProdOfLocals.cartabs_setOfTuples

  (* Compute the abstraction of a difference between two sets of states. 
     Return the difference in the first component. Return a saturated version 
     of the second set of states in the second component. *)
  let abstract_setOfStates_Minus_setOfStates (s1:setOfStates) (ex:setOfStates):abstractElement*setOfStates = 
    (* Debug.print_debug ("AbstractConcreteElement.abstract_setOfStates_Minus_setOfStates started on "^(string_of_setOfStates s1)^" and "^(string_of_setOfStates ex)^".\n"); *)
    let f scp1 = function
	Some scp2 ->
	  let (cp,new2)=CartProdOfLocals.cartabs_diff_setOfTuples_setOfTuples scp1 scp2 in ((if CartProdOfLocals.is_empty cp then None else Some cp), Some new2)
      |	None -> ((Some (fst (CartProdOfLocals.cartabs_diff_setOfTuples_setOfTuples scp1 CartProdOfLocals.empty_setOfTuples))),None)
    in let result=MapFromSharedStates.diff_apply_change s1 ex f in
      (* Debug.print_debug ("AbstractConcreteElement.abstract_setOfStates_Minus_setOfStates finished with ("^(string_of_abstractElement (fst result))^","^(string_of_setOfStates (snd result))^").\n"); *)
      result


  (* creates a join of two abstract elements *)
  let join_abstract_elements ae1 ae2=
    (* Debug.print_debug ("AbstractConcreteElement.join_abstract_elements started on "^(string_of_abstractElement ae1)^" and "^(string_of_abstractElement ae2)^".\n"); *)
    let result=MapFromSharedStates.union (CartProdOfLocals.join) ae1 ae2 in
      (* Debug.print_debug ("AbstractConcreteElement.join_abstract_elements finished with "^(string_of_abstractElement result)^".\n"); *)
      result

  (* tests whether an abstract element is less than or equal to another 
     abstract element. *)
  let le_abstract_elements ae1 ae2=
    (* Debug.print_debug ("AbstractConcreteElement.le_abstract_elements started on "^(string_of_abstractElement ae1)^" and "^(string_of_abstractElement ae2)^".\n"); *)
    let result= MapFromSharedStates.subset_apply CartProdOfLocals.subset ae1 ae2 in
      (* Debug.print_debug ("AbstractConcreteElement.le_abstract_elements finished with "^(string_of_bool result)^".\n"); *)
      result 


  (* concretize an abstract element and unite with a set of states. *)
  let concretize_abstractElement_union_setOfStates ae sos=
    (* Debug.print_debug ("AbstractConcreteElement.concretize_abstractElement_union_setOfStates started on "^(string_of_abstractElement ae)^".\n"); *)
    let result=MapFromSharedStates.union CartProdOfLocals.union_setOfTuples_setOfTuples (MapFromSharedStates.map CartProdOfLocals.setOfTuples_from_CP ae) sos in
      (* Debug.print_debug "AbstractConcreteElement.concretize_abstractElement_union_setOfStates finished.\n"; *)
      result

  (* [join_abstract_element_abstract_setOfStates_minus_setOfStates a a' e] 
     computes the join of a with (cartesianAbstraction(a'\e)) in the first 
     component. A more saturated version of e is returned in the second 
     component. *)
  let join_abstract_element_cartabs_setOfStates_minus_setOfStates (a:abstractElement) (a':setOfStates) (e:setOfStates) : abstractElement*setOfStates =
(*
    Debug.print_debug ("AbstractConcreteElement.join_abstract_element_cartabs_setOfStates_minus_setOfStates started on a="^(string_of_abstractElement a)^", a'="^(string_of_setOfStates a')^", e="^(string_of_setOfStates e)^".\n");
    let result=*)
    MapFromSharedStates.union_apply3
      (fun (c:CartProdOfLocals.t option) (sos:CartProdOfLocals.setOfTuples) (e:CartProdOfLocals.setOfTuples option) -> 
(*	 Debug.print_debug ("internal function called with c="^(match c with None -> "None" | Some cp -> CartProdOfLocals.string_of_CP cp)^" and sos="^(CartProdOfLocals.string_of_setOfTuples sos)^" and e="^(match e with None->"None"|Some ee ->CartProdOfLocals.string_of_setOfTuples ee)^".\n"); 
	 let result2= *)
	 let cp= match c with Some cp ->cp | None -> CartProdOfLocals.empty and
	     esot = match e with Some esot -> esot| None ->CartProdOfLocals.empty_setOfTuples in
	 let ((res:CartProdOfLocals.t),(new_esot:CartProdOfLocals.setOfTuples))=(CartProdOfLocals.join_CP_cartabs_diff_setOfTuples_setOfTuples cp sos esot)
	 in ((if CartProdOfLocals.is_empty res then None else Some res),(match e with None -> None | Some _ -> Some new_esot))
(*	 in Debug.print_debug ("internal function finished with ("^(match fst result2 with None -> "None" | Some res2 -> CartProdOfLocals.string_of_CP res2)^", "^(match snd result2 with None -> "None" | Some n ->CartProdOfLocals.string_of_setOfTuples n)^").\n"); result2 *)

      )
      a a' e
(*    in (Debug.print_debug ("AbstractConcreteElement.join_abstract_element_cartabs_setOfStates_minus_setOfStates finished with result ("^(string_of_abstractElement (fst result))^","^(string_of_setOfStates (snd result))^").\n")); result *)

  (* Gets raised if a precondition for generating an exception set is 
     not satisfied. *)
  exception ExceptionSetNotFound

  (* exception_set c b f where c is an abstract element and b and f are
     sets of states finds some subset e of b such that the intersection of f 
     with the cartesian abstraction of (c union (b\e)) is empty. It assumes 
     that the intersection of f with (c union b) is empty. *)
  let exception_set c b f = 
    (* Debug.print_debug ("AbstractConcreteElement.exception_set started on abs elem "^(string_of_abstractElement c)^", concrete set "^(string_of_setOfStates b)^" and bad states "^(string_of_setOfStates f)^".\n"); *)
    let result=
      MapFromSharedStates.fold
	(fun g sot_of_b sofar_e ->
	   let cp_of_c= try MapFromSharedStates.find g c with Not_found -> CartProdOfLocals.empty
	   in try
	       let sot_of_f=MapFromSharedStates.find g f in
	       let u=CartProdOfLocals.union_CP_setOfTuples cp_of_c sot_of_b in
	       let au_sot=CartProdOfLocals.setOfTuples_from_CP (CartProdOfLocals.cartabs_setOfTuples u) in
	       let i=CartProdOfLocals.inter_setOfTuples_setOfTuples au_sot sot_of_f in
		 if CartProdOfLocals.is_empty_setOfTuples i then sofar_e else
(*
		   MapFromSharedStates.add g (CartProdOfLocals.exception_set_for_setOfTuples cp_of_c sot_of_b sot_of_f) sofar_e 
*)
		   MapFromSharedStates.add g (CartProdOfLocals.eager_exception_set_for_setOfTuples cp_of_c sot_of_b sot_of_f) sofar_e

	     with Not_found -> sofar_e
	       | CartProdOfLocals.ExceptionSetNotFound -> raise ExceptionSetNotFound
	) b MapFromSharedStates.empty 
    in
     (* Debug.print_debug ("AbstractConcreteElement.exception_set finished with "^(string_of_setOfStates result)^".\n"); *)
      result

  (* [no_mutex_setOfStates inside all globs n] creates the set of states
      { (g,x1,...,xn) | g in globs and exists i unequal j: xi in inside 
      and xj in inside and for all k unequal i,j we have xk in all}. *)
  let no_mutex_setOfStates (inside:setOfComponent) (all:setOfComponent) (globs:setOfShared) (n:int):setOfStates =
    let no_mutex_sos=CartProdOfLocals.no_mutex_setOfTuples inside all n in
      SetOfShared.fold 
	(fun g sofar -> MapFromSharedStates.add g no_mutex_sos sofar) 
	globs empty_setOfStates

  (* [no_assertion_setOfStates inside all globs n] creates the set of states
     { (g,x1,...,xn) | g in globs and exists i : xi in inside 
     and for all k unequal i we have xk in all}. *)
  let no_assertion_setOfStates (inside:setOfComponent) (all:setOfComponent) (globs:setOfShared) (n:int):setOfStates =
    let sos=CartProdOfLocals.mutex_setOfTuples inside all n in
      SetOfShared.fold 
	(fun g sofar -> MapFromSharedStates.add g sos sofar) 
	globs empty_setOfStates

  (* create a string representation of a list of states *)
  let string_of_trace t = 
    "["^(
      let lst=List.map string_of_setOfStates t in String.concat "," lst
    )^"]"
end
