(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

module type PairOfOrderedTypes = sig
  type left_type
  type right_type
  val compare_left: left_type -> left_type ->int
  val compare_right:  right_type -> right_type ->int
end

module type S = sig
  type left_key
  type right_key
  type t (* type of bijective maps from left_key to right_key*)
  val empty: t
  val is_empty: t -> bool
  val add: left_key -> right_key -> t -> t (* add x y m returns a map containing the same bindings as m, plus a binding of x to y. If x or y were already bound in m, their previous bindings disappear. *)
  val find_left: left_key -> t -> right_key (* find x m returns m(x), if x is in the domain of m, or raises Not_found if no such binding exists. *)
  val find_right: right_key -> t -> left_key (* find y m returns m^-1(y), or raises Not_found if no such binding exists. *)
  val mem_left: left_key -> t -> bool (** [mem_left x m] returns true if there is pair in [m] with [x] in the first component *)
  val mem_right: right_key -> t -> bool (** [mem_right y m] returns true if there is pair in [m] with [y] in the second component *)
  val max_left: t->left_key (** Return the greatest element of the domain (with respect to the [P.compare_left] ordering), or raise Not_found if the bijection is empty. *)
  val max_right: t->right_key (** Return the greatest element of the range (with respect to the [P.compare_right] ordering), or raise Not_found if the bijection is empty. *)
end

module Make (P:PairOfOrderedTypes): (S with type left_key=P.left_type and type right_key=P.right_type) =
  struct
    type left_key = P.left_type
    type right_key = P.right_type
    module OrderedLeftKey= struct
      type t = left_key
      let compare = P.compare_left
    end
    module OrderedRightKey = struct
      type t = right_key
      let compare = P.compare_right
    end
    module LeftMap = ExtMap.Make (OrderedLeftKey)
    module RightMap = ExtMap.Make (OrderedRightKey)
    type leftMap = right_key LeftMap.t
    type rightMap = left_key RightMap.t
    type t = (leftMap)*(rightMap)
    let empty = (LeftMap.empty,RightMap.empty)
    let is_empty ((lm,rm):t) = LeftMap.is_empty lm
    let add (l:left_key) (r:right_key) ((lm,rm):t) :t = 
      ((try
	  let rmr=RightMap.find r rm
	  in LeftMap.add l r (LeftMap.remove rmr lm)
	with Not_found -> LeftMap.add l r lm)
	 ,
       (try 
	  let lml=LeftMap.find l lm 
	  in RightMap.add r l (RightMap.remove lml rm)
	with Not_found -> RightMap.add r l rm))
	
    let find_left (l:left_key) ((lm,rm):t): right_key = LeftMap.find l lm
      
    let find_right (r:right_key) ((lm,rm):t) :left_key = RightMap.find r rm

    let mem_left (x:left_key) ((lm,rm):t) = LeftMap.mem x lm

    let mem_right (y:right_key) ((lm,rm):t) = RightMap.mem y rm

    let max_left ((lm,rm):t):left_key =LeftMap.max_key lm

    let max_right ((lm,rm):t):right_key = RightMap.max_key rm
  end
