(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(*  Alexander Malkis. Added function max_key.                          *)
(***********************************************************************)

(* $Id: map.ml,v 1.17 2005/08/13 20:59:37 doligez Exp $ *)
(* $Id: extmap.ml,v 1.18 2008/03/07 20:59:37 alexmalk Exp $ *)


module type OrderedType =
  sig
    type t
    val compare: t -> t -> int
  end

module type S =
  sig
    type key
    type +'a t
    val empty: 'a t
    val is_empty: 'a t -> bool
    val is_singleton: 'a t -> bool
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val mem:  key -> 'a t -> bool
    val iter: (key -> 'a -> unit) -> 'a t -> unit
    val map: ('a -> 'b) -> 'a t -> 'b t
    val mapi: (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val compare: ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal: ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
    val min_binding: 'a t->key*'a
    val max_key: 'a t->key 
    val for_all: 'a t->(key->'a->bool)->bool
    val exists: 'a t->(key->'a->bool)->(key*'a)
    val any_binding: 'a t ->(key*'a)
    val intersect_apply: 'a t -> 'a t -> ('a -> 'a -> ('a option)) -> 'a t
    val diff_apply: ('a t) -> ('a t) -> ('a -> 'a -> ('a option)) -> ('a t)
    val diff_apply_change: ('a t)->('a t)->('a -> 'a option -> ('b option *'a option))->(('b t)*('a t))
    val union: ('a ->'a ->'a)->'a t->'a t->'a t
    val union_apply3: ('a option -> 'b ->'c option ->(('a option)*('c option)))->('a t)->('b t)->('c t)->(('a t)*('c t))
    val subset_apply: ('a->'a->bool)->'a t->'a t->bool
    val cardinal: 'a t -> int
end

module Make(Ord: OrderedType) = struct

    type key = Ord.t

    type 'a t =
        Empty
      | Node of 'a t * key * 'a * 'a t * int

    let height = function
        Empty -> 0
      | Node(_,_,_,_,h) -> h

    let create l x d r =
      let hl = height l and hr = height r in
      Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

    let bal l x d r =
      let hl = match l with Empty -> 0 | Node(_,_,_,_,h) -> h in
      let hr = match r with Empty -> 0 | Node(_,_,_,_,h) -> h in
      if hl > hr + 2 then begin
        match l with
          Empty -> invalid_arg "ExtMap.bal"
        | Node(ll, lv, ld, lr, _) ->
            if height ll >= height lr then
              create ll lv ld (create lr x d r)
            else begin
              match lr with
                Empty -> invalid_arg "ExtMap.bal"
              | Node(lrl, lrv, lrd, lrr, _)->
                  create (create ll lv ld lrl) lrv lrd (create lrr x d r)
            end
      end else if hr > hl + 2 then begin
        match r with
          Empty -> invalid_arg "ExtMap.bal"
        | Node(rl, rv, rd, rr, _) ->
            if height rr >= height rl then
              create (create l x d rl) rv rd rr
            else begin
              match rl with
                Empty -> invalid_arg "ExtMap.bal"
              | Node(rll, rlv, rld, rlr, _) ->
                  create (create l x d rll) rlv rld (create rlr rv rd rr)
            end
      end else
        Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

    let empty = Empty

    let is_empty = function Empty -> true | _ -> false

    let is_singleton = function
      	Node(Empty, _, _, Empty, _) -> true
      |	_ ->false

    let rec add x data = function
        Empty ->
          Node(Empty, x, data, Empty, 1)
      | Node(l, v, d, r, h) ->
          let c = Ord.compare x v in
          if c = 0 then
            Node(l, x, data, r, h)
          else if c < 0 then
            bal (add x data l) v d r
          else
            bal l v d (add x data r)

    let rec find x = function
        Empty ->
          raise Not_found
      | Node(l, v, d, r, _) ->
          let c = Ord.compare x v in
          if c = 0 then d
          else find x (if c < 0 then l else r)

    let rec mem x = function
        Empty ->
          false
      | Node(l, v, d, r, _) ->
          let c = Ord.compare x v in
          c = 0 || mem x (if c < 0 then l else r)

    let rec min_binding = function
        Empty -> raise Not_found
      | Node(Empty, x, d, r, _) -> (x, d)
      | Node(l, x, d, r, _) -> min_binding l

    let rec remove_min_binding = function
        Empty -> invalid_arg "ExtMap.remove_min_binding"
      | Node(Empty, x, d, r, _) -> r
      | Node(l, x, d, r, _) -> bal (remove_min_binding l) x d r

    let merge t1 t2 =
      match (t1, t2) with
        (Empty, t) -> t
      | (t, Empty) -> t
      | (_, _) ->
          let (x, d) = min_binding t2 in
          bal t1 x d (remove_min_binding t2)

    let rec remove x = function
        Empty ->
          Empty
      | Node(l, v, d, r, h) ->
          let c = Ord.compare x v in
          if c = 0 then
            merge l r
          else if c < 0 then
            bal (remove x l) v d r
          else
            bal l v d (remove x r)

    let rec iter f = function
        Empty -> ()
      | Node(l, v, d, r, _) ->
          iter f l; f v d; iter f r

    let rec map f = function
        Empty               -> Empty
      | Node(l, v, d, r, h) -> Node(map f l, v, f d, map f r, h)

    let rec mapi f = function
        Empty               -> Empty
      | Node(l, v, d, r, h) -> Node(mapi f l, v, f v d, mapi f r, h)

    let rec fold f m accu =
      match m with
        Empty -> accu
      | Node(l, v, d, r, _) ->
          fold f r (f v d (fold f l accu))

    type 'a enumeration = End | More of key * 'a * 'a t * 'a enumeration

    let rec cons_enum m e =
      match m with
        Empty -> e
      | Node(l, v, d, r, _) -> cons_enum l (More(v, d, r, e))

    let compare cmp m1 m2 =
      let rec compare_aux e1 e2 =
          match (e1, e2) with
          (End, End) -> 0
        | (End, _)  -> -1
        | (_, End) -> 1
        | (More(v1, d1, r1, e1), More(v2, d2, r2, e2)) ->
            let c = Ord.compare v1 v2 in
            if c <> 0 then c else
            let c = cmp d1 d2 in
            if c <> 0 then c else
            compare_aux (cons_enum r1 e1) (cons_enum r2 e2)
      in compare_aux (cons_enum m1 End) (cons_enum m2 End)

    let equal cmp m1 m2 =
      let rec equal_aux e1 e2 =
          match (e1, e2) with
          (End, End) -> true
        | (End, _)  -> false
        | (_, End) -> false
        | (More(v1, d1, r1, e1), More(v2, d2, r2, e2)) ->
            Ord.compare v1 v2 = 0 && cmp d1 d2 &&
            equal_aux (cons_enum r1 e1) (cons_enum r2 e2)
      in equal_aux (cons_enum m1 End) (cons_enum m2 End)

    let rec max_key (m:'a t): key = 
      match m with
       Empty -> raise Not_found
     | Node(l, x, d, Empty, _) -> x
     | Node(l, x, d, r, _) -> max_key r

    let rec for_all (m:'a t) (f:key->'a->bool) =
      match m with
        Empty -> true
      | Node(l, x, d, r, _) -> ((f x d) && (for_all l f) && (for_all r f))

    let exists m f =
      let rec exists_aux =function
	  Empty -> raise Not_found
	| Node(l,x,d,r,_) -> 
	    if f x d then (x,d) else 
	      try exists_aux l
	      with Not_found -> exists_aux r
      in exists_aux m

    let any_binding (m:'a t):(key*'a) =
      match m with
        Empty -> raise Not_found
      | Node(_,x,d,_,_) -> (x,d)

    (* type used for addressing inside trees. *)
    type childType=LeftChild|RightChild|Root

    (* [intersect_apply m1 m2 f] creates a new map with bindings 
       {(v,k)| ex. k1,k2: (v,k1)in m1 and (v,k2) in m2 and Some k = f k1 k2 }*)
    let intersect_apply m1 m2 f =
      (* find_minimal_with_path find a node with a minimal key inside a map 
	 and returns a path to it, which is a list of maps from the bottom to 
	 the top *)
      let rec find_minimal_with_path m= match m with
	  [] -> raise Not_found
	| h::t -> (match h with
		       (Empty,_) -> raise Not_found
		     | ((Node(Empty,x,d,r,_)),_) -> m
		     | ((Node(l,x,d,r,_)),_) -> find_minimal_with_path ((l,LeftChild)::m))
	    
      in (* gives a key of the current path given by m*)
      let key_of_path:(('a t*childType)list)->key  = function
	  [] -> raise Not_found
	| h::t -> (match h with
		       (Empty,_) -> raise Not_found
		     | ((Node(_,x,_,_,_)),_) -> x)
	      
      (* retunrs the key of the current node *)
(*    in
      let key_of_node = function
	  (Empty,_) -> 
	    raise (Invalid_argument "ExtMap.key_of_node got an empty node")
	| (Node(_,x,_,_,_),_) -> x
*)	    
      in (* gives the data of the current path given by m*)
      let data_of_path = function
	  [] -> raise Not_found
	| h::t -> (match h with
		       (Empty,_) -> raise Not_found
		     | (Node(_,_,d,_,_),_) ->d)
	    
      in (* next_right_ascendant mlst searches for the lowest node on the 
	    path from mlst to the root so that the path goes through its 
	    left branch. If no such node is found, Not_found is raised *)
      let rec next_right_ascendant:(('a t*childType) list) -> (('a t*childType) list) = function
	  [] -> raise Not_found
	| [h] -> raise Not_found
	| h::t -> 
	    (match h with
		 (_,LeftChild) -> t
	       | (_,RightChild) -> next_right_ascendant t
	       | (_,Root) -> raise (Invalid_argument "ExtMap.next_right_ascendant is given an invalid path with a root in the middle."))

      in (* next_larger_than x m searches for a smallest key k>x in the tree 
	    of a path given by m. Searches only the portion of the tree below 
	    m and with the keys greater than that of m. Raises Not_found if no 
	    such found, otherwise returns the path to the node.*)
      let rec next_greater_than (x:key) (mlst:('a t*childType) list) : ('a t*childType) list= 
	let rec ngt_inside_map ml1 = match ml1 with
	    [] -> raise Not_found
	  | m::t -> 
	      (match m with
		   (Empty,_) -> raise Not_found
		 | ((Node(l,k,d,r,_)),_) ->
		     if (Ord.compare k x)<=0 
		     then ngt_inside_map ((r,RightChild)::ml1)
		     else (try ngt_inside_map ((l,LeftChild)::ml1) with Not_found -> ml1))
	in (try ngt_inside_map mlst
	    with Not_found -> 
	      (let p:('a t*childType) list=(next_right_ascendant mlst)
	       in if ((Ord.compare (key_of_path p) x)<=0) 
		 then next_greater_than x p else p))
	     
      in (* next_greater_equal x m searches for a smallest key k>=x in the 
	    tree of a path given by m. Searches only the portion of the tree 
	    below m and with the keys greater than that of m. Raises Not_found 
	    if no such found, otherwise returns the path to the node.*)
      let rec next_greater_equal x mlst =
	let rec nge_inside_map ml1 = match ml1 with
	    [] -> raise Not_found
	  | m::t ->
	      (match m with
		   (Empty,_) -> raise Not_found
		 | ((Node(l,k,d,r,_)),_) ->
		     let comp=Ord.compare k x in
		       if comp<0 then nge_inside_map ((r,RightChild)::ml1)
		       else if comp=0 then ml1
		       else try nge_inside_map ((l,LeftChild)::ml1) with Not_found -> ml1)
	in match mlst with
	    [] -> raise Not_found
	  | h::t -> (try nge_inside_map mlst
	             with Not_found -> 
		       let p=next_right_ascendant mlst
		       in if (Ord.compare (key_of_path p) x)>=0 then p 
			 else next_greater_equal x p)
	
      in (* add_new_values_with_path min1 min2 m checks whether min1 and min2 
	    point to the equal keys. If yes, it adds the key together with the 
	    value computed by f to m and calls itself recursively on the pair 
	    pointing to the next largest elements. If min1-key is smaller than 
	    the key of min2, it searches for the smallest key in the tree of 
	    min1 that is greater than or equal to the key of min2. If an equal 
	    key is found, the corresponding key is added and the procedure is 
	    called recusively on the pair pointing to the next largest 
	    elements. If a greater key is found, the procedure is called 
	    recursively on the new pair. 
	    If the key of min1 is larger than the key of min2, it searches for 
	    the smallest key inside the tree of min2 that is greater than or 
	    equal to the key pointing to min1. If an equal key is found, the 
	    procedure adds the key together with the value given by f and 
	    proceeds to the next largest pair. If a greater key is found, the 
	    procedure is called recursively on the new pair. *)
      let rec add_new_values_with_path (m1:('a t*childType)list) (m2:('a t*childType)list) m =
	match (m1,m2) with
	    ([],_)-> m
	  | (_,[])->m
	  | (_,_) ->
	      let key1=key_of_path m1 and key2=key_of_path m2
	      in let comp=Ord.compare key1 key2 in
		if comp=0 then
		  let d1=data_of_path m1 and d2=data_of_path m2 in 
		  let new_m=
		    (match f d1 d2 with
			 None -> m
		       | Some(data) -> add key1 data m)
		  in
		    (try 
		       let n1=next_greater_than key1 m1 
		       and n2=next_greater_than key2 m2
		       in add_new_values_with_path n1 n2 new_m
		     with  Not_found -> new_m)
		else if comp>0 then
		  (try 
		     let n2=next_greater_equal key1 m2
		     in add_new_values_with_path m1 n2 m
		   with Not_found -> m)
		else 
		  (try
		     let n1=next_greater_equal key2 m1
		     in add_new_values_with_path n1 m2 m
		   with Not_found -> m)
		    
    in try (let min1=find_minimal_with_path [(m1,Root)] 
	    and min2=find_minimal_with_path [(m2,Root)]
	    in add_new_values_with_path min1 min2 Empty)
	with Not_found -> Empty

    (* Creates a balanced map with contents l union r union {(x,d)}, assuming
       that all keys of l are less than x and all keys of r are greater than x.
    *)
    let rec join l x d r = 
      match (l,r) with
	  (Empty,_) -> add x d r
	| (_,Empty) -> add x d l
	| (Node(ll,lx,ld,lr,lh),Node(rl, rx,rd, rr, rh)) ->
	    if lh>rh+2 then bal ll lx ld (join lr x d r)  else
	      if rh>lh+2 then bal (join l x d rl) rx rd rr else
		create l x d r

    (* Merge two trees l and r into one.
       All elements of l must precede the elements of r.
       No assumption on the heights of l and r. *)
   let concat t1 t2 =
      match (t1, t2) with
        (Empty, t) -> t
      | (t, Empty) -> t
      | (_, _) -> 
	  let (x,d)=min_binding t2 in join t1 x d (remove_min_binding t2)

   (* Splitting.  split x s returns a triple (l, present, r) where
      - l is the set of elements of s that are < x
      - r is the set of elements of s that are > x
      - present is None if s contains no element with key x,
      or Some d if s contains an element with key x and data d. *)
   let rec split (x:key) (s:'a t) : ('a t*'a option*'a t)= match s with
        Empty ->
          (Empty, None, Empty)
      | Node(l, k,v, r, _) ->
          let c = Ord.compare x k in
          if c = 0 then (l, Some v, r)
          else if c < 0 then
            let (ll, pres, rl) = split x l in (ll, pres, (join rl k v r))
          else
            let (lr, pres, rr) = split x r in ((join l k v lr), pres, rr)


   (* diff_apply s1 s2 f computes a difference between maps s1 and s2,
      applying each time a function f on matching keys. More precisely,
      it computes 
      {(k,v) | (k,v)in s1 and k is not a key of s2
             or
               (k,v1)in s1 and (k,v2)in s2 and f v1 v2 = Some v
      }
   *)
   let diff_apply (s1: 'a t) (s2: 'a t) (f: 'a -> 'a -> ('a option)): ('a t) =
     let rec diff_apply_f s1 s2 = match (s1,s2) with
	 (Empty,t2) -> Empty
       | (t1,Empty) -> t1
       | (Node(l1,x1,d1,r1,_),t2) ->
	   let (l2,ns_d2,r2)=split x1 t2 in
	   let left_diff=diff_apply_f l1 l2 in
	   let right_diff=diff_apply_f r1 r2 in
	     match ns_d2 with
		 None -> concat left_diff right_diff
	       | Some d2 -> 
		   (match f d1 d2 with
			None -> concat left_diff right_diff
		      | Some d -> join left_diff x1 d right_diff)
     in diff_apply_f s1 s2

   (* given a map, apply f to all its values. If the result is Some dd, 
      store (x,dd) in the output map. *)
   let rec map_optional f =function
       Empty -> Empty
     | Node(l,x,d,r,_) -> 
	 let left=map_optional f l and right=map_optional f r in
	   (match f d with
		None -> concat left right
	      | Some dd -> join left x dd right)

   (* Computes a difference between maps s1 and s2. If a key occurs only in 
      s1, the binding is left "as is" in the difference. If a key k occurs 
      both in s1 and s2 with data d1 and d2, and if fst(f d1 d2) = Some d, 
      then (k,d) is also a binding of the difference. If a key does not occur 
      in s1 or if a key occurs in s1 with data d1 and in s2 with data d2, but 
      fst(f d1 d2)=None, then the key doesn't belong to the difference. The 
      difference is returned as the first component of the function. 
      Additionally, a rebuild of the map s2 is returned in the second 
      component. A rebuild of s2 is a map s2' on the same set of keys so 
      that for all keys k of s2 (or s2') we have: whenever (k,d2) is in s2 and 
      (k,d2') in s2', then either d2=d2' or there is some d1 with (k,d1) in s1
      and snd(f d1 (Some d2))=Some d2'; moreover, if snd(f d1 (Some d2))=None,
      then d2=d2'. *)
    let rec diff_apply_change (s1:'a t) (s2:'a t) (f:'a -> 'a option -> ('b option *'a option)):(('b t)*('a t)) =
      match (s1, s2) with
        (Empty, t2) -> (Empty,t2)
      | (t1, Empty) -> ((map_optional (fun x->fst(f x None)) t1),Empty)
      | (Node(l1, x1, d1, r1, _), t2) ->
	  let ((l2:'a t),(ns_d2:'a option),(r2:'a t))=split x1 t2 in
	  let (left_diff,left_new_s2) = diff_apply_change l1 l2 f and
	      (right_diff,right_new_s2)=diff_apply_change r1 r2 f in 
	  let (d_diff,d_new_s2)=f d1 ns_d2 in
	    ((match d_diff with
		 None -> concat left_diff right_diff
	       | Some d -> join left_diff x1 d right_diff),
	     (match (ns_d2,d_new_s2) with
		  (None,None) -> concat left_new_s2 right_new_s2
		| (None,Some _) -> raise (Invalid_argument "ExtMap.diff_apply_change was called with a function that accepts None as the second argument but returns Some _ as the second argument.\n")
		| (Some d2,None) -> join left_new_s2 x1 d2 right_new_s2
		| (Some d2,Some d2') -> join left_new_s2 x1 d2' right_new_s2))

    (* [union s1 s2 f] creates a map containing 
       {(k,d)| (k,d1)in s1 and k is not a key of s2 and d1=d} 
       union
       {(k,d)| (k,d2)in s2 and k is not a key of s1 and d2=d} 
       union
       {(k,d) | (k,d1)in s1 and (k,d2)in s2 and f d1 d2 = d}. *)
    let union f s1 s2=
      let rec un s1 s2=
	match (s1, s2) with
            (Empty, t2) -> t2
	  | (t1, Empty) -> t1
	  | (Node(l1, k1, d1, r1, h1), Node(l2, k2, d2, r2, h2)) ->
	      if h1>=h2 then let (l2, pres, r2) = split k1 s2 in
		(match pres with
		     None ->  (join (un l1 l2) k1 d1 (un r1 r2))
		   | Some d2 -> join (un l1 l2) k1 (f d1 d2) (un r1 r2))
	    else let (l1,pres,r1)=split k2 s1 in
	      (match pres with
		   None -> join (un l1 l2) k2 d2 (un r1 r2)
		 | Some d1 -> join (un l1 l2) k2 (f d1 d2) (un r1 r2))
      in un s1 s2


    (* [union_apply3 f s1 s2 s3] creates the map
       {(k,d)| (k,d1)in s1 and k is not a key of s2 and d1 = d} 
       union
       {(k,d)| (k,d2)in s2 and k is not a key of s1 and 
       ((k,d3) is in s3 and fst(f None d2 (Some d3)) = Some d)
       or 
       (k is not a key of s3 and fst(f None d2 None) = Some d)}
       }
       union
       {(k,d)| (k,d1)in s1 and (k,d2)in s2 and 
       ((k,d3) is in s3 and fst (f (Some d1) d2 (Some d3)) = Some d)
       or 
       (k is not a key of s3 and fst(f (Some d1) d2 None) = Some d)}. 
       This map is returned in the first component.
       Additionally, a rebuild of the map s3 is returned in the second 
       component. A rebuild of s3 is a map s3' on the same set of keys so 
       that for all keys k of s3 (or s3') we have: whenever (k,d3) is in s3 
       and (k,d3') in s3', then either d3=d3' or there is some d2 with (k,d2) 
       in s2 and snd(f _ d2 (Some d3))=Some d3'. *)

    let union_apply3 (f:'a option -> 'b ->'c option ->(('a option)*('c option))) (s1:'a t) (s2:'b t) (s3:'c t) :(('a t)*('c t)) =
      let rec apply2 (s2:'b t) (s3:'c t):('a t)*('c t) =
	match (s2,s3) with 
	    (Empty,t3)->(Empty,t3)
	  | (t2,Empty)-> 
	      ((map_optional (fun (d2:'b) -> fst(f None d2 None)) t2),Empty)
 	  | (Node(l2, k2, d2, r2, h2), _) ->
	      let (l3,pres3,r3)=split k2 s3 in 
	      let (d,nd3)=f None d2 pres3 and
		  (left,newl3)=apply2 l2 l3 and
		  (right,newr3)=apply2 r2 r3 in
		((match d with
		      None -> concat left right
		    | Some dd ->join left k2 dd right),
		 (match (pres3,nd3) with 
		      (None,_) -> concat newl3 newr3
		    | (Some d3,None) -> join newl3 k2 d3 newr3
		    | (Some _,Some d3) -> join newl3 k2 d3 newr3))
      in  
      let rec un (s1:'a t) (s2:'b t) (s3:'c t):(('a t)*('c t)) =
	match (s1,s2) with
	    (Empty, t2) -> apply2 s2 s3
	  | (t1,Empty)->(t1,s3)
	  | (Node(l1, k1, d1, r1, h1), Node(l2, k2, d2, r2, h2)) ->
	      if h1>=h2 then
		let (l2,pres2,r2)=split k1 s2 and (l3,pres3,r3)=split k1 s3 in
		let (left,newl3)=un l1 l2 l3 and (right,newr3)=un r1 r2 r3 in
		  match pres2 with
		      None -> ((join left k1 d1 right),
			       (match pres3 with 
				    None -> concat newl3 newr3
				  | Some d3 -> join newl3 k1 d3 newr3))
		    | Some d2 -> 
			let (d,nd3)=f (Some d1) d2 pres3 in
			  ((match d with
				None -> concat left right
			      | Some dd -> join left k1 dd right),
			   (match (pres3,nd3) with 
				(None,_) -> concat newl3 newr3
			      | (Some d3,None) -> join newl3 k1 d3 newr3
			      | (Some _,Some d3) -> join newl3 k1 d3 newr3))
	      else
		let (l1,pres1,r1)=split k2 s1 and (l3,pres3,r3)=split k2 s3 in
		let (left,newl3)=un l1 l2 l3 and (right,newr3)=un r1 r2 r3 in
		let (d,nd3)=f pres1 d2 pres3 in
		  ((match d with
			None -> concat left right
		      | Some dd -> join left k2 dd right),
		   (match (pres3,nd3) with 
			(None,_) -> concat newl3 newr3
		      | (Some d3,None) -> join newl3 k2 d3 newr3
		      | (Some _,Some d3)->join newl3 k2 d3 newr3))
      in un s1 s2 s3


    (* subset function checks whether each key of s1 is a key of s2 and 
       whenever (k,d1) is a binding of s1 and (k,d2) of s2, f d1 d2 holds *)  
    let subset_apply (f:'a->'a->bool) (s1:'a t) (s2:'a t):bool =
      let rec subset_aux s1 s2 =
	match (s1,s2) with
	    (Empty,_) -> true
	  | (_,Empty)->false
	  | (Node(l1,k1,d1,r1,_),Node(l2,k2,d2,r2,_)) ->
	      let c=Ord.compare k1 k2 in
		if c=0 then ((f d1 d2)&&(subset_aux l1 l2)&&(subset_aux r1 r2))
		else if c<0 then
		  (subset_aux (Node(l1,k1,d1,Empty,0)) l2) &&
		    (subset_aux r1 s2)
		else (subset_aux (Node(Empty,k1,d1,r1,0)) r2) &&
		  (subset_aux l1 s2)
      in subset_aux s1 s2

    (* counts the number of bindings *)
    let rec cardinal = function
        Empty -> 0
      | Node(l,_,_,r,_) -> cardinal l + 1 + cardinal r
	  
end
