(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

open ModelChecker

(*
We will check the program of 2 threads, one producer and one consumer.

Let N>0, M>0 be constant.

Shared variables are A[0..M-1], B[0..M-1], buffer[0..N-1], inpointer, outpointer, count, Bufman.
Local variables are program counter, x and i for producer, y and j for consumer.

The initial values of variables are:
bufman = 0 (the lock is not held by anyone)
A[0],...,A[M-1] arbitrary boolean
B[0],...,B[M-1] arbitrary boolean
buffer[0],...,buffer[N-1] arbitrary boolean
inpointer=outpointer=count=0
x,y arbitrary

The used functions below are f = g = negation

Producer:
A: if i>=M goto J;
B:   x:=g(A[i]);
C:   atomic{ await (Bufman==0 && count<N); Bufman:=tid; }
D:     buffer[inpointer]:=x;
E:     inpointer:=inpointer+1 mod N;
F:     count++;
G:   Bufman:=0;
H:   i++;
I:   goto A;
J:

Consumer:
K: if j>=M goto T;
L:   atomic{ await (Bufman==0 && count>0); Bufman:=tid; }
M:     y:=buffer[outpointer];
N:     outpointer:=outpointer+1 mod N;
O:     count--;
P:   Bufman:=0;
Q:   B[j]:=f(y);
R:   j++;
S:   goto K:
T:

Problem: the algorithm can't be easily adapted to many producers/consumers.

The property to be verified that at the end of execution, i.e. when the threads are at J or T, B[k]=f(g(A[k])) for all k = 1,...,M.

A shared state is a bitvector
Bufman_1 Bufman_0 A[0]...A[M-1] buffer[0]..buffer[N-1] B[0]...B[M-1] inpointer_{floor(log(N-1))}...inpointer_0 outpointer_{floor(log(N-1))}...outpointer_0 count_{floor(log(N))}...count_0

A local state is a bitvector
i_{floor(log(M))}...i_0 pc_4...pc_0 (for producers)
or
j_{floor(log(M))}...j_0 pc_4...pc_0 (for consumers)
*)

(* States. *)
module OrderedInt=struct
  type t=int
  let compare=Pervasives.compare
end

module MapFromInts=ExtMap.Make(OrderedInt)

type pc=PC_A|PC_B|PC_C|PC_D|PC_E|PC_F|PC_G|PC_H|PC_I|PC_J|PC_K|PC_L|PC_M|PC_N|PC_O|PC_P|PC_Q|PC_R|PC_S|PC_T

(* Produces a local state of any threads. Takes a value of i (or j), then x(or y) and a program counter and returns a pair of a numerical value and a string. *)

let local_state var value pc :AC.component= (((var*64)+(if value then 32 else 0)+(match pc with PC_A->0 | PC_B ->1 | PC_C -> 2 | PC_D ->3 | PC_E -> 4 | PC_F -> 5 | PC_G -> 6 | PC_H ->7 | PC_I ->8 | PC_J ->9 | PC_K ->10 | PC_L ->11 | PC_M ->12 | PC_N -> 13 | PC_O ->14 | PC_P ->15 | PC_Q ->16 | PC_R ->17 | PC_S -> 18 | PC_T -> 19)),
			  ((string_of_int var)^(string_of_bool value)^(match pc with PC_A -> "A"  | PC_B -> "B"  | PC_C -> "C"  | PC_D -> "D"  | PC_E -> "E"  | PC_F -> "F"  | PC_G -> "G"  | PC_H -> "H"  | PC_I -> "I"  | PC_J -> "J"  | PC_K -> "K"  | PC_L -> "L"  | PC_M -> "M"  | PC_N -> "N"  | PC_O -> "O"  | PC_P -> "P"  | PC_Q -> "Q"  | PC_R -> "R"  | PC_S -> "S"  | PC_T -> "T")))


let string_of_bool_array = (Array.fold_left (fun sofar elt -> ((if elt then "T" else "F")^sofar)) "")

let shared_state m n bufman a buffer b inpointer outpointer count =
  let floor_log2 number = int_of_float ((log (float number))/.(log 2.0)) in
  let offset_count=0 in
  let offset_outpointer=offset_count+(floor_log2 n)+1 in
  let offset_inpointer=offset_outpointer +(floor_log2 (n-1))+1 in
  let offset_B=offset_inpointer+(floor_log2 (n-1))+1 in
  let offset_buffer=offset_B+m in
  let offset_A=offset_buffer+n in
  let offset_bufman=offset_A+m in
  let intOfA = Array.fold_right (fun arr_elt sofar -> ((2*sofar)+(if arr_elt then 1 else 0))) a 0 in
  let intOfB = Array.fold_right (fun arr_elt sofar -> 2*sofar+(if arr_elt then 1 else 0)) b 0 in
  let intOfBuffer = Array.fold_right (fun arr_elt sofar -> 2*sofar+(if arr_elt then 1 else 0)) buffer 0 in
  (((bufman lsl offset_bufman)+(intOfA lsl offset_A)+(intOfBuffer lsl offset_buffer)+(intOfB lsl offset_B)+(inpointer lsl offset_inpointer)+(outpointer lsl offset_outpointer)+(count lsl offset_count)),("("^(string_of_int bufman)^","^(string_of_bool_array a)^","^(string_of_bool_array buffer)^","^(string_of_bool_array b)^","^(string_of_int inpointer)^","^(string_of_int outpointer)^","^(string_of_int count)^")"))

exception Done

let xor b1 b2 = (b1 && (not b2)) || ((not b1) && b2)

let next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x) =
  if not x then (bufman,a,buffer,b,inpointer,outpointer,count,i,true)
  else if i<m then (bufman,a,buffer,b,inpointer,outpointer,count,i+1,false)
  else if count<n then (bufman,a,buffer,b,inpointer,outpointer,(count+1),0,false)
  else if outpointer<(n-1) then (bufman,a,buffer,b,inpointer,(outpointer+1),0,0,false)
  else if inpointer<(n-1) then  (bufman,a,buffer,b,(inpointer+1),0,0,0,false)
  else let next_larger_array arr = 
    let (new_lst,overfl)=Array.fold_left (fun (sofar,overfl) elt -> (((xor elt overfl)::sofar),(elt && overfl))) ([],false) arr in
      ((Array.of_list new_lst),overfl)
  in let (new_b,overfl_b)=next_larger_array b in
    if not overfl_b then (bufman,a,buffer,new_b,0,0,0,0,false)
    else let (new_buffer,overfl_buffer)=next_larger_array buffer in
      if not overfl_buffer then (bufman,a,new_buffer,new_b,0,0,0,0,false)
      else let (new_a,overfl_a)=next_larger_array a in
	if not overfl_a then (bufman,new_a,new_buffer,new_b,0,0,0,0,false)
	else if bufman<2 then ((bufman+1),new_a,new_buffer,new_b,0,0,0,0,false)
	else raise Done

(* [plus_bounded_counter x y] gives x+1, if x+1<y, or x, if x+1>=y. *)
let plus_bounded_counter x y = if (x+1)<y then x+1 else x

(* [minus_bounded_counter x] gives x-1, if x-1>=0, or x, if x<=0. *)
let minus_bounded_counter x = if x>0 then x-1 else x

(* [plus_modulo_counter x y] return a nonnegative modulus (x+1) mod y *)
let rec plus_modulo_counter x y = if x<0 then plus_modulo_counter (x+y) y else (x+1) mod y

(* [minus_modulo_counter x y] returns a nonnegative modulus (x-1) mod y *)
let rec minus_modulo_counter x y = if x<0 then minus_modulo_counter (x+y) y else (x-1) mod y

let funf x = not x
let fung = funf

let array_of_false len =Array.make len false

(* returns the transition relation of the threads *)
let create_post m n =
  let init_thread_state = (0,(array_of_false m),(array_of_false n),(array_of_false m),0,0,0,0,false) in
  let create_post_producer sofar = (
    let rec create_post_a (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t= if i>=m then
	AC.single_transition g (local_state i x PC_A) 0 g (local_state i x PC_J)
      else
	AC.single_transition g (local_state i x PC_A) 0 g (local_state i x PC_B)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_a (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_b (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_B) 0 g (local_state i (fung (a.(i))) PC_C)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_b (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_c (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n 0 a buffer b inpointer outpointer count in
      let g'=shared_state m n 1 a buffer b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_C) 0 g' (local_state i x PC_D)
      in let sofar'=if ((bufman==0) && (count<n)) then AC.add_transition_transitionSet t sofar else sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_c (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_d (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let buffer'=Array.copy buffer in 
      let _=buffer'.(inpointer)<-x in
      let g'=shared_state m n bufman a buffer' b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_D) 0 g' (local_state i x PC_E)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_d (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_e (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b (plus_modulo_counter inpointer n) outpointer count in
      let t= AC.single_transition g (local_state i x PC_E) 0 g' (local_state i x PC_F)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_e (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_f (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b inpointer outpointer (plus_bounded_counter count (n+1)) in
      let t= AC.single_transition g (local_state i x PC_F) 0 g' (local_state i x PC_G)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_f (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_g (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n 0 a buffer b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_G) 0 g' (local_state i x PC_H)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_g (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_h (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_H) 0 g' (local_state (plus_bounded_counter i (m+1)) x PC_I)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_h (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in
    let rec create_post_i (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b inpointer outpointer count in
      let t= AC.single_transition g (local_state i x PC_I) 0 g' (local_state i x PC_A)
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,i,x) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,i,x)
	in create_post_i (bufman,a,buffer,b,inpointer,outpointer,count,i,x) sofar'
	with Done -> sofar'
    in 
      create_post_a init_thread_state 
	(create_post_b init_thread_state 
	   (create_post_c init_thread_state 
	      (create_post_d init_thread_state 
		 (create_post_e init_thread_state 
		    (create_post_f init_thread_state 
		       (create_post_g init_thread_state 
			  (create_post_h init_thread_state 
			     (create_post_i init_thread_state sofar))))))))
  )
  in
  let create_post_consumer sofar = (
    let rec create_post_k (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_K) 1 g (local_state j y (if j>=m then PC_T else PC_L))
      in let sofar'=AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_k (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_l (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n 0 a buffer b inpointer outpointer count in
      let g'=shared_state m n 2 a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_L) 1 g' (local_state j y PC_M)
      in let sofar'=if count>0 then AC.add_transition_transitionSet t sofar else sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_l (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_m (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_M) 1 g (local_state j (buffer.(outpointer)) PC_N)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_m (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_n (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b inpointer (plus_modulo_counter outpointer n) count in
      let t=AC.single_transition g (local_state j y PC_N) 1 g' (local_state j y PC_O)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_m (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_o (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n bufman a buffer b inpointer outpointer (minus_bounded_counter count) in
      let t=AC.single_transition g (local_state j y PC_O) 1 g' (local_state j y PC_P)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_o (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_p (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let g'=shared_state m n 0 a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_P) 1 g' (local_state j y PC_Q)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_p (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_q (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let b'=Array.copy b in
      let _ =b'.(j)<-(funf y) in
      let g'=shared_state m n bufman a buffer b' inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_Q) 1 g' (local_state j y PC_R)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_q (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_r (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_R) 1 g (local_state (plus_bounded_counter j (m+1)) y PC_S)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_r (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
    let rec create_post_s (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar =
      let g=shared_state m n bufman a buffer b inpointer outpointer count in
      let t=AC.single_transition g (local_state j y PC_S) 1 g (local_state j y PC_K)
      in let sofar'= AC.add_transition_transitionSet t sofar in
	try let (bufman,a,buffer,b,inpointer,outpointer,count,j,y) = next_thread_state m n (bufman,a,buffer,b,inpointer,outpointer,count,j,y)
	in create_post_r (bufman,a,buffer,b,inpointer,outpointer,count,j,y) sofar'
	with Done -> sofar'
    in
      create_post_k init_thread_state 
	(create_post_l init_thread_state 
	   (create_post_m init_thread_state 
	      (create_post_n init_thread_state 
		 (create_post_o init_thread_state 
		    (create_post_p init_thread_state 
		       (create_post_q init_thread_state 
			  (create_post_r init_thread_state 
			     (create_post_s init_thread_state sofar))))))))
  )
  in
    create_post_producer (create_post_consumer AC.empty_transitionSet)

let next_shared_state m n (bufman,a,buffer,b,inpointer,outpointer,count) =
  if count<n then (bufman,a,buffer,b,inpointer,outpointer,(count+1))
  else if outpointer<(n-1) then (bufman,a,buffer,b,inpointer,(outpointer+1),0)
  else if inpointer<(n-1) then  (bufman,a,buffer,b,(inpointer+1),0,0)
  else let next_larger_array arr = 
    let (new_lst,overfl)=Array.fold_left (fun (sofar,overfl) elt -> (((xor elt overfl)::sofar),(elt && overfl))) ([],false) arr in
      ((Array.of_list new_lst),overfl)
  in let (new_b,overfl_b)=next_larger_array b in
    if not overfl_b then (bufman,a,buffer,new_b,0,0,0)
    else let (new_buffer,overfl_buffer)=next_larger_array buffer in
      if not overfl_buffer then (bufman,a,new_buffer,new_b,0,0,0)
      else let (new_a,overfl_a)=next_larger_array a in
	if not overfl_a then (bufman,new_a,new_buffer,new_b,0,0,0)
	else if bufman<2 then ((bufman+1),new_a,new_buffer,new_b,0,0,0)
	else raise Done


module SetOfShared=Set.Make(AbstractConcreteElement.MakeOrderedShared(SL))

(* The invariant to be proven is that at the end of computation, a=b. *)
let error_glob m n =
  let rec error_glob_aux sofar (bufman,a,buffer,b,inpointer,outpointer,count) =
    let sofar'=if a<>b then SetOfShared.add (shared_state m n bufman a buffer b inpointer outpointer count) sofar else sofar in
      try error_glob_aux sofar' (next_shared_state m n (bufman,a,buffer,b,inpointer,outpointer,count))
      with Done -> sofar
  in error_glob_aux SetOfShared.empty (0,(array_of_false m),(array_of_false n),(array_of_false m),0,0,0)

module SetOfComponent=Set.Make(AbstractConcreteElement.MakeOrderedLocal(SL))

let last_local_producer m n =
  let rec last_local_producer_aux sofar i x =
    let new_local=local_state i x PC_J in
    let sofar'=SetOfComponent.add new_local sofar in
    if not x then 
      last_local_producer_aux sofar' i true 
    else if i<m then last_local_producer_aux sofar' (i+1) false else sofar'
  in last_local_producer_aux SetOfComponent.empty 0 false

let last_local_consumer m n =
  let rec last_local_consumer_aux sofar j y =
    let new_local=local_state j y PC_T in
    let sofar'=SetOfComponent.add new_local sofar in
    if not y then 
      last_local_consumer_aux sofar' j true 
    else if j<m then last_local_consumer_aux sofar' (j+1) false else sofar'
  in last_local_consumer_aux SetOfComponent.empty 0 false

let create_error m n =
  let gs = error_glob m n in
  let lp = last_local_producer m n in
  let lc = last_local_consumer m n in
    AC.no_assertion_setOfStates lp lc gs 2

let create_init m n :AC.setOfStates =
  let locals=Array.of_list [local_state 0 false PC_A; local_state 0 false PC_K] in
  let rec init_aux sofar (bufman,a,buffer,b,inpointer,outpointer,count) =(
    let sofar'=if (inpointer=0 && outpointer=0 && count=0 && bufman=0) then AC.add_state_setOfStates (AC.single_state (shared_state m n bufman a buffer b inpointer outpointer count) locals) sofar else sofar in
    let ((bufman',a',buffer',b',inpointer',outpointer',count'),dne)=(
      try ((next_shared_state m n (bufman,a,buffer,b,inpointer,outpointer,count)),false)
      with Done -> ((bufman,a,buffer,b,inpointer,outpointer,count),true))
    in if dne then sofar' else init_aux sofar' (bufman',a',buffer',b',inpointer',outpointer',count')
  )
  in init_aux AC.empty_setOfStates (0,(array_of_false m),(array_of_false n),(array_of_false m),0,0,0)

let start_model_checker m n =
  let init=create_init m n in  Debug.print_debug "Init created.\n";
  let error=create_error m n in Debug.print_debug "Error created.\n";
  let post=create_post m n in Debug.print_debug "Post created.\n";
  let mc_result=model_check init error post in
    match mc_result with
        Safe invariant ->print_string ("Safe. Invariant:\n"^(AC.string_of_setOfStates invariant)^".\n")
      | Unsafe trace -> print_string ("Unsafe. Error trace:\n"^(AC.string_of_trace trace)^".\n")

let main () =
  Debug.print_debug "Starting...\n";
  let error_message ()=print_string "Checking the producer/consumer example (Owicki example 4.2). Syntax:\nOwicki-4.2-producer-consumer m n\nwhere m and n are at least 2.\n" in
  if ((Array.length(Sys.argv))<>3) then error_message () else
    try 
      let m=int_of_string Sys.argv.(1) in
	if m<2 then raise (Failure "m too small") else
	  let n=int_of_string Sys.argv.(2) in
	    if n<2 then raise (Failure "n too small") else 
	      start_model_checker m n
    with Failure f -> (print_string ("Failure: "^f^"\n"); error_message ())
;;
main ()

