(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

open ModelChecker

(*
A program consists of r readers and w writers, each of them containing m critical sections.

Shared variables: ar, aw, ww, where 0<=ww<=w, 0<=aw<=1, 0<=ar<=r.
Initially ar=aw=ww=0.

Reader code:
A0: assume 0==ww; ar++;
B0: ar--;
A1: assume 0==ww; ar++;
B1: ar--;
...
A(m-1): assume 0==ww; ar++;
B(m-1): ar--; goto A0;

Writer code:
A0: ww++;
B0: assume 0==ar && 0==aw; aw++;
C0: aw--; ww--;
A1: ww++;
B1: assume 0==ar && 0==aw; aw++;
C1: aw--; ww--;
...
A(m-1): ww++;
B(m-1): assume 0==ar && 0==aw; aw++;
C(m-1): aw--; ww--; goto A0;

The property to be proven is that no more than one writer is at Ci for any i.
*)

type lcst = A | B | C

(* creates a local state out of its letter part and the number of critical section *)
let local_state (lc:lcst) (critsec:int) =
 match lc with
     A -> ((3*critsec),("A"^(string_of_int critsec)))
   | B -> ((3*critsec+1),("B"^(string_of_int critsec)))
   | C -> ((3*critsec+2),("C"^(string_of_int critsec)))

(* a shared state is a bitvector
ar_{floor(log2 r)}..ar_0 aw ww_{floor(log2 w)}..ww_0
*)

let rec no_of_bits_for_number n =
  if n<=1 then 1 else 1+(no_of_bits_for_number (n lsr 1))

(* gives "0" or "1" for false or true, respectively *)
let int_string_of_bool (b:bool) : string = if b then "1" else "0"

(* gives 0 or 1 for false or true, respectively *)
let int_of_bool (b:bool) : int = if b then 1 else 0

let shared_state no_of_bits_for_w ar aw ww :AC.shared =
  let offset_aw = no_of_bits_for_w in
  let offset_ar = offset_aw+1 in
  let value=(ar lsl offset_ar)+((int_of_bool aw) lsl offset_aw) + ww in
    (value,("("^(string_of_int ar)^","^(int_string_of_bool aw)^","^(string_of_int ww)^")"))

(* the second component returns true on overflow *)
let next_shared_state r w (ar,aw,ww)=
  if ww<w then ((ar,aw,(ww+1)),false)
  else if not aw then ((ar,true,0),false)
  else if ar<r then (((ar+1),false,0),false) else ((0,false,0),true)

(* [plus_bounded_counter x y] gives x+1, if x+1<y, or x, if x+1>=y. *)
let plus_bounded_counter x y = if (x+1)<y then x+1 else x

(* [minus_bounded_counter x] gives x-1, if x-1>=0, or x, if x<=0. *)
let minus_bounded_counter x = if x>0 then x-1 else x

(* [plus_modulo_counter x y] return a nonnegative modulus (x+1) mod y *)
let rec plus_modulo_counter x y = if x<0 then plus_modulo_counter (x+y) y else (x+1) mod y

(* [minus_modulo_counter x y] returns a nonnegative modulus (x-1) mod y *)
let rec minus_modulo_counter x y = if x<0 then minus_modulo_counter (x+y) y else (x-1) mod y

let create_post r w m no_of_bits_for_w=
  let rec create_post_reader i sofar =
    if i>=r then sofar else
      let rec create_post_reader_a j (ar,aw,ww) sofar =
        if j>=m then sofar else
	  let sofar'=
	    if ww=0 then
	      let g = shared_state no_of_bits_for_w ar aw ww in
	      let g' = shared_state no_of_bits_for_w (plus_bounded_counter ar (r+1)) aw ww in
	      let t=AC.single_transition g (local_state A j) i g' (local_state B j) in
		AC.add_transition_transitionSet t sofar
	    else sofar
	  in let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	    if overfl then create_post_reader_a (j+1) (0,false,0) sofar'
	    else create_post_reader_a j (ar',aw',ww') sofar'
      in
      let rec create_post_reader_b j (ar,aw,ww) sofar =
	if j>=m then sofar else
	  let g= shared_state no_of_bits_for_w ar aw ww in
	  let g'=shared_state no_of_bits_for_w (minus_bounded_counter ar) aw ww in
	  let l=local_state B j in
	  let l'=local_state A (plus_modulo_counter j m) in
	  let t=AC.single_transition g l i g' l' in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	  let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	    if overfl then create_post_reader_b (j+1) (0,false,0) sofar'
	    else create_post_reader_b j (ar',aw',ww') sofar'
      in 
      let sofar'=create_post_reader_a 0 (0,false,0) (create_post_reader_b 0 (0,false,0) sofar) in 
	create_post_reader (i+1) sofar'
  in 
  let rplusw=r+w in
  let rec create_post_writer i sofar =
    if i>=rplusw then sofar else
      let rec create_post_writer_a j (ar,aw,ww) sofar =
	if j>=m then sofar else
	  let g= shared_state no_of_bits_for_w ar aw ww in
	  let g'=shared_state no_of_bits_for_w ar aw (plus_bounded_counter ww (w+1)) in
	  let t=AC.single_transition g (local_state A j) i g' (local_state B j) in
	  let sofar'=  AC.add_transition_transitionSet t sofar in
	  let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
	    if overfl then create_post_writer_a (j+1) (0,false,0) sofar'
	    else create_post_writer_a j (ar',aw',ww') sofar'
      in
      let rec create_post_writer_b j (ar,aw,ww) sofar =
	if j>=m then sofar else
	  let sofar'=
	    if ar=0 && not aw then 
	      let g= shared_state no_of_bits_for_w ar aw ww in
	      let g'=shared_state no_of_bits_for_w ar true ww in
	      let t=AC.single_transition g (local_state B j) i g' (local_state C j) in
		AC.add_transition_transitionSet t sofar
	    else sofar
	  in let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
	  if overfl then create_post_writer_b (j+1) (0,false,0) sofar' 
	  else create_post_writer_b j (ar',aw',ww') sofar'
      in
      let rec create_post_writer_c j (ar,aw,ww) sofar =
	if j>=m then sofar else
	  let g= shared_state no_of_bits_for_w ar aw ww in
	  let g'=shared_state no_of_bits_for_w ar false (minus_bounded_counter ww) in
	  let l=local_state C j in
	  let l'=local_state A (plus_modulo_counter j m) in
	  let t=AC.single_transition g l i g' l' in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	  let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	    if overfl then (create_post_writer_c (j+1) (0,false,0) sofar') else (create_post_writer_c j (ar',aw',ww') sofar')
      in
      let pwc=create_post_writer_c 0 (0,false,0) sofar in
      let pwbc =create_post_writer_b 0 (0,false,0) pwc in
      let sofar'=create_post_writer_a 0 (0,false,0) pwbc 
      in create_post_writer (i+1) sofar'
  in 
  let writer_post=(create_post_writer r (AC.empty_transitionSet))
  in create_post_reader 0 writer_post

let create_init r w no_of_bits_for_w =
  AC.singleton_setOfStates (shared_state (no_of_bits_for_number w) 0 false 0) (Array.make (r+w) (local_state A 0))

module SetOfComponent=Set.Make(AbstractConcreteElement.MakeOrderedLocal(SL))
module SetOfShared=Set.Make(AbstractConcreteElement.MakeOrderedShared(SL))

let create_error r w m no_of_bits_for_w=
  let rec create_all_shared_aux sofar (ar,aw,ww) =
    let sofar'=SetOfShared.add (shared_state no_of_bits_for_w ar aw ww) sofar in
    let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
      if overfl then sofar' else create_all_shared_aux sofar' (ar',aw',ww')
  in let globs=create_all_shared_aux SetOfShared.empty (0,false,0) in
  let rec create_all_locals_aux sofar letter j =
    match letter with
	A -> if j>=m then create_all_locals_aux sofar B 0 else
	  let l=local_state A j in
	    create_all_locals_aux (SetOfComponent.add l sofar) A (j+1)
      | B -> if j>=m then create_all_locals_aux sofar C 0 else
	  let l=local_state B j in
	    create_all_locals_aux (SetOfComponent.add l sofar) B (j+1)
      | C -> if j>=m then sofar else
	  let l=local_state C j in
	    create_all_locals_aux (SetOfComponent.add l sofar) C (j+1)
  in let all_locals=create_all_locals_aux (SetOfComponent.empty) A 0 in
  let rec create_all_inside_aux sofar j =
    if j>=m then sofar else
      let l=local_state C j in
	create_all_inside_aux (SetOfComponent.add l sofar) (j+1)
  in let inside_locals=create_all_inside_aux SetOfComponent.empty 0
  in AC.no_mutex_setOfStates inside_locals all_locals globs (r+w)

let start_model_checker r w m =
(*  let tr_start_time=Unix.time () in*)
  let no_of_bits_for_w = no_of_bits_for_number w in
  let init =create_init r w no_of_bits_for_w in
  let post=create_post r w m no_of_bits_for_w in
  let error=create_error r w m no_of_bits_for_w in
(*  let tr_end_time=Unix.time () in
  let _= (print_string ("Transition system created in "^(string_of_int (int_of_float (tr_end_time-.tr_start_time)))^" s.\n"); flush stdout) in
  let mc_start_time=Unix.time () in *)
  let mc_result=model_check init error post in
(*  let mc_finish_time=Unix.time () in
    print_string ("Model checking done in "^(string_of_int (int_of_float (mc_finish_time-.mc_start_time)))^" s.\n";);*)
    match mc_result with
	Safe invariant -> print_string ("Safe. Invariant:\n"^(AC.string_of_setOfStates invariant)^".\n")
      | Unsafe trace -> print_string ("Unsafe. Error trace:\n"^(AC.string_of_trace trace)^".\n")

let main () =
  let error_message ()=print_string "Checking the readers-writers from Owicki's thesis, Fig. 5.4. Syntax:\nOwicki-5.4-readers-writers <number_of_readers> <number of writers> <number_of_critical_sections_per_thread>\n" in
    if ((Array.length(Sys.argv))<>4) then (print_string "Error: supply three arguments.\n"; error_message ()) else
      try
	let r=int_of_string Sys.argv.(1) in
	  if r<0 then raise (Failure "The number of readers should be nonnegative!") else
	    let w=int_of_string Sys.argv.(2) in
	      if w<0 then raise (Failure "The number of writers should be nonnegative!")	else 
		let m=int_of_string Sys.argv.(3) in
		  if m<0 then raise (Failure "The number of critical sections should be nonnegative!")	else 
		    start_model_checker r w m
      with Failure f -> (print_string ("Error: "^f^"\n"); error_message ())
;;

main ()

