(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

open ModelChecker

(*
A program consists of r readers and w writers.

Shared variables: ar, aw, ww, where 0<=ww<=w, 0<=aw<=w, 0<=ar<=r.
Initially ar=aw=ww=0.

Reader code:
A: assume 0==ww; ar++;
B: ar--; goto A;

Writer code:
A: ww++;
B: assume 0==ar && 0==aw; aw++;
C: aw--; ww--; goto A;

The property to be proven is that no more than one writer is at C.

*)

let local_a=(0,"A")
let local_b=(1,"B")
let local_c=(2,"C")

(* a shared state is a bitvector
ar_{floor(log2 r)}..ar_0 aw_{floor(log2 w)}..aw_0 ww_{floor(log2 w)}..ww_0
*)

let rec no_of_bits_for_number n =
  if n<=1 then 1 else 1+(no_of_bits_for_number (n lsr 1))

let shared_state w ar aw ww :AC.shared =
(*
  let floor_log2 number = int_of_float ((log (float number))/.(log 2.0)) in
  let floor_log8 =int_of_float ((log 8.0)/.(log 2.0)) in
  let floor_log2_w_plus1 =1+(floor_log2 w) in
*)
  let no_of_bits_for_w= no_of_bits_for_number w in
  let offset_aw = no_of_bits_for_w in
  let offset_ar = offset_aw +no_of_bits_for_w in
  let value=(ar lsl offset_ar)+(aw lsl offset_aw) + ww in
    (value,("("^(string_of_int ar)^","^(string_of_int aw)^","^(string_of_int ww)^")"))

(* the second component returns true on overflow *)
let next_shared_state r w (ar,aw,ww)=
  if ww<w then ((ar,aw,(ww+1)),false)
  else if aw<w then ((ar,(aw+1),0),false)
  else if ar<r then (((ar+1),0,0),false) else ((0,0,0),true)



(* [plus_bounded_counter x y] gives x+1, if x+1<y, or x, if x+1>=y. *)
let plus_bounded_counter x y = if (x+1)<y then x+1 else x

(* [minus_bounded_counter x] gives x-1, if x-1>=0, or x, if x<=0. *)
let minus_bounded_counter x = if x>0 then x-1 else x

(* [plus_modulo_counter x y] return a nonnegative modulus (x+1) mod y *)
let rec plus_modulo_counter x y = if x<0 then plus_modulo_counter (x+y) y else (x+1) mod y

(* [minus_modulo_counter x y] returns a nonnegative modulus (x-1) mod y *)
let rec minus_modulo_counter x y = if x<0 then minus_modulo_counter (x+y) y else (x-1) mod y

let create_post r w =
  let rec create_post_reader i sofar =
    if i>=r then sofar else
      let rec create_post_reader_a (ar,aw,ww) sofar =
	let sofar'=
	  if ww=0 then
	    let g = shared_state w ar aw ww in
	    let g' = shared_state w (plus_bounded_counter ar (r+1)) aw ww in
	    let t=AC.single_transition g local_a i g' local_b in
	      AC.add_transition_transitionSet t sofar
	  else sofar
	in let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	  if overfl then sofar' else create_post_reader_a (ar',aw',ww') sofar'
      in
      let rec create_post_reader_b (ar,aw,ww) sofar =
	let g= shared_state w ar aw ww in
	let g'=shared_state w (minus_bounded_counter ar) aw ww in
	let t=AC.single_transition g local_b i g' local_a in
	let sofar'=AC.add_transition_transitionSet t sofar in
	let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	  if overfl then sofar' else create_post_reader_b (ar',aw',ww') sofar'
      in 
      let sofar'=create_post_reader_a (0,0,0) (create_post_reader_b (0,0,0) sofar) in
	create_post_reader (i+1) sofar'
  in 
  let rplusw=r+w in
  let rec create_post_writer i sofar =
    if i>=rplusw then sofar else
      let rec create_post_writer_a (ar,aw,ww) sofar =
	let g= shared_state w ar aw ww in
	let g'=shared_state w ar aw (plus_bounded_counter ww (w+1)) in
	let t=AC.single_transition g local_a i g' local_b in
	let sofar'=  AC.add_transition_transitionSet t sofar in
	let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
	  if overfl then sofar' else create_post_writer_a (ar',aw',ww') sofar'
      in
      let rec create_post_writer_b (ar,aw,ww) sofar =
	let sofar'=
	  if ar=0 && aw=0 then 
	    let g= shared_state w ar aw ww in
	    let g'=shared_state w ar (plus_bounded_counter aw (w+1)) ww in
	    let t=AC.single_transition g local_b i g' local_c in
	      AC.add_transition_transitionSet t sofar
	  else sofar
	in let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
	  if overfl then sofar' else create_post_writer_b (ar',aw',ww') sofar'
      in
      let rec create_post_writer_c (ar,aw,ww) sofar =
	let g= shared_state w ar aw ww in
	let g'=shared_state w ar (minus_bounded_counter aw) (minus_bounded_counter ww) in
	let t=AC.single_transition g local_c i g' local_a in
	let sofar'=AC.add_transition_transitionSet t sofar in
	let ((ar',aw',ww'),overfl) = next_shared_state r w (ar,aw,ww) in
	  if overfl then sofar' else create_post_writer_c (ar',aw',ww') sofar'
      in
      let sofar'=create_post_writer_a (0,0,0) (create_post_writer_b (0,0,0) (create_post_writer_c (0,0,0) sofar)) 
      in create_post_writer (i+1) sofar'
  in create_post_reader 0 (create_post_writer r (AC.empty_transitionSet))

let create_init r w =
  AC.singleton_setOfStates (shared_state w 0 0 0) (Array.make (r+w) local_a)

module SetOfComponent=Set.Make(AbstractConcreteElement.MakeOrderedLocal(SL))
module SetOfShared=Set.Make(AbstractConcreteElement.MakeOrderedShared(SL))

let create_error r w =
  let rec create_all_shared_aux sofar (ar,aw,ww) =
    let sofar'=SetOfShared.add (shared_state w ar aw ww) sofar in
    let ((ar',aw',ww'),overfl)=next_shared_state r w (ar,aw,ww) in
      if overfl then sofar' else create_all_shared_aux sofar' (ar',aw',ww')
  in let globs=create_all_shared_aux SetOfShared.empty (0,0,0) in
  let inside=SetOfComponent.singleton local_c in
  let all=SetOfComponent.add local_a (SetOfComponent.add local_b (SetOfComponent.singleton local_c)) in
    AC.no_mutex_setOfStates inside all globs (r+w)

let start_model_checker r w =
  let init =create_init r w in
  let post=create_post r w in
  let error=create_error r w in
  let mc_result=model_check init error post in
    match mc_result with
	Safe invariant -> print_string ("Safe. Invariant:\n"^(AC.string_of_setOfStates invariant)^".\n")
      | Unsafe trace -> print_string ("Unsafe. Error trace:\n"^(AC.string_of_trace trace)^".\n")

let main () =
  let error_message ()=print_string "Checking the readers-writers from Owicki's thesis, Fig. 5.4. Syntax:\nOwicki-5.4-readers-writers <number_of_readers> <number of writers>\n" in
    if ((Array.length(Sys.argv))<>3) then (print_string "Error: supply two arguments.\n"; error_message ()) else
      try
	let r=int_of_string Sys.argv.(1) in
	  if r<0 then raise (Failure "The number of readers should be nonnegative!") else
	    let w=int_of_string Sys.argv.(2) in
	      if w<0 then raise (Failure "The number of writers should be nonnegative!")	else start_model_checker r w
      with Failure f -> (print_string ("Error: "^f^"\n"); error_message ())
;;

main ()

