open ModelChecker

(* We will check the program of n+1 threads, where thread 0 is
A: stoppingFlag:=1;
B: pendingIO--;
C: if(!pendingIO)
D:   stoppingEvent:=1;
E: assume stoppingEvent;
F: stopped:=true;
G:

and threads 1 to n are

H: pendingIO++;
I: if(!stoppingFlag) {
J:   assert(!stopped); /*do work*/ }
K: pendingIO--;
L: if(!pendingIO)
M:   stoppingEvent:=1;
N:

The initial state is pendingIO=1, stoppingFlag=stoppingEvent=stopped=false, pc0=A, pc1=...=pcn=H.

The numerical values of local states are
A=0
B=1
C=2
D=3
E=4
F=5
G=6
H=7
I=8
J=9
K=10
L=11
M=12
N=13

A shared state is a bitvectors
a_{ceil(log(n+1))} ... a_0 stoppingFlag stoppingEvent stopped
*)

(* States. *)
module OrderedInt=struct
  type t=int
  let compare=Pervasives.compare
end

module MapFromInts=ExtMap.Make(OrderedInt)


let local_a = (0,"A")
let local_b = (1,"B")
let local_c = (2,"C")
let local_d = (3,"D")
let local_e = (4,"E")
let local_f = (5,"F")
let local_g = (6,"G")
let local_h = (7,"H")
let local_i = (8,"I")
let local_j = (9,"J")
let local_k = (10,"K")
let local_l = (11,"L")
let local_m = (12,"M")
let local_n = (13,"N")

(* boolean convertions *)
let int_of_bool b = if b then 1 else 0
let string_of_bool b = if b then "T" else "F"

(* constructs a shared state (i.e. an integer together with its string) from the pendingIO, stoppingFlag, stoppingEvent, stopped. *)
let shared_state (pendingIO:int) (stoppingFlag:bool) (stoppingEvent:bool) (stopped:bool) = 
  ((pendingIO*8)+((int_of_bool stoppingFlag)*4)+((int_of_bool stoppingEvent)*2)+(int_of_bool stopped),
 (string_of_int pendingIO)^(string_of_bool stoppingFlag)^(string_of_bool stoppingEvent)^(string_of_bool stopped))

module SetOfComponent=Set.Make(AbstractConcreteElement.MakeOrderedLocal(SL))
module SetOfShared=Set.Make(AbstractConcreteElement.MakeOrderedShared(SL))

exception Done

(* counter of length 3. throws Done on overflow *)
let three_counter (x,y,z)=
  match (x,y,z) with
      (_,_,false) -> (x,y,true)
    | (_,false,_) -> (x,true,false)
    | (false,_,_) -> (true,false,false)
    | (true,true,true) -> raise Done

(* minus_counter x m gives x-1, if x-1>=0, or m, if x<0. *)
let minus_counter x m = if x>0 then x-1 else m

(* plus_counter x m gives x+1, if x+1<=m, or 0, if x+1>m. *)
let plus_counter x m = if (x+1)<=m then x+1 else 0
  
(* take every pio modulo n+2 *)
let max_pio_of_thread_number n = n+2

(* creates the set of transitions. *)
let create_post (n:int) :AC.transitionSet =
  let max_pio=max_pio_of_thread_number n in
  let create_post_thread0 (sofar:AC.transitionSet) = (
    let rec create_post_a pio sF sE st sofar= (
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF sE st in
	let g'=shared_state pio true sE st in
	let t=AC.single_transition g local_a 0 g' local_b in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_a pio sF' sE' st' sofar'
	  with Done ->
	    create_post_a (pio+1) false false false sofar')
    in 
    let rec create_post_b pio sF sE st sofar =(
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF sE st in
	let g'=shared_state (minus_counter pio max_pio) sF sE st in
	let t=AC.single_transition g local_b 0 g' local_c in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_b pio sF' sE' st' sofar'
	  with Done ->
	    create_post_b (pio+1) false false false sofar')
    in
    let rec create_post_c pio sF sE st sofar =(
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF sE st in
	let g'=g in
	let t=AC.single_transition g local_c 0 g' (if pio=0 then local_d else local_e) in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_c pio sF' sE' st' sofar'
	  with Done ->
	    create_post_c (pio+1) false false false sofar')
    in
    let rec create_post_d pio sF sE st sofar =(
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF sE st in
	let g'=shared_state pio sF true st in
	let t=AC.single_transition g local_d 0 g' local_e in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_d pio sF' sE' st' sofar'
	  with Done ->
	    create_post_d (pio+1) false false false sofar')
    in
    let rec create_post_e pio sF sE st sofar =(
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF true st in
	let g'=shared_state pio sF true st in
	let t=AC.single_transition g local_e 0 g' local_f in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_e pio sF' sE' st' sofar'
	  with Done ->
	    create_post_e (pio+1) false false false sofar')
    in
    let rec create_post_f pio sF sE st sofar =(
      if pio>max_pio then sofar 
      else 
	let g=shared_state pio sF sE st in
	let g'=shared_state pio sF sE true in
	let t=AC.single_transition g local_f 0 g' local_g in
	let sofar'=AC.add_transition_transitionSet t sofar in
	  try let (sF',sE',st')=three_counter (sF,sE,st) in
	    create_post_f pio sF' sE' st' sofar'
	  with Done ->
	    create_post_f (pio+1) false false false sofar')
    in 
      create_post_a 0 false false false 
	(create_post_b 0 false false false
	   (create_post_c 0 false false false
	      (create_post_d 0 false false false
		 (create_post_e 0 false false false
		    (create_post_f 0 false false false sofar)))))
  )
  in
  let rec create_post_thread (i:int) (sofar:AC.transitionSet) = (
    if i>n then sofar else
      let rec create_post_h pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE st in
	  let g'=shared_state (plus_counter pio max_pio) sF sE st in
	  let t=AC.single_transition g local_h i g' local_i in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_h pio sF' sE' st' sofar'
	    with Done ->
	      create_post_h (pio+1) false false false sofar')
      in
      let rec create_post_i pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE st in
	  let g'=g in
	  let t=AC.single_transition g local_i i g' (if sF then local_k else local_j) in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_i pio sF' sE' st' sofar'
	    with Done ->
	      create_post_i (pio+1) false false false sofar')
      in
      let rec create_post_j pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE false in
	  let g'=g in
	  let t=AC.single_transition g local_j i g' local_k in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_j pio sF' sE' st' sofar'
	    with Done ->
	      create_post_j (pio+1) false false false sofar')
      in
      let rec create_post_k pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE st in
	  let g'=shared_state (minus_counter pio max_pio) sF sE st in
	  let t=AC.single_transition g local_k i g' local_l in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_k pio sF' sE' st' sofar'
	    with Done ->
	      create_post_k (pio+1) false false false sofar')
      in
      let rec create_post_l pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE st in
	  let g'=shared_state pio sF sE st in
	  let t=AC.single_transition g local_l i g' (if pio=0 then local_m else local_n) in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_l pio sF' sE' st' sofar'
	    with Done ->
	      create_post_l (pio+1) false false false sofar')
      in
      let rec create_post_m pio sF sE st sofar= (
	if pio>max_pio then sofar 
	else 
	  let g=shared_state pio sF sE st in
	  let g'=shared_state pio sF true st in
	  let t=AC.single_transition g local_m i g' local_n in
	  let sofar'=AC.add_transition_transitionSet t sofar in
	    try let (sF',sE',st')=three_counter (sF,sE,st) in
	      create_post_m pio sF' sE' st' sofar'
	    with Done ->
	      create_post_m (pio+1) false false false sofar')
      in
	create_post_h 0 false false false 
	  (create_post_i 0 false false false
	     (create_post_j 0 false false false
		(create_post_k 0 false false false
		   (create_post_l 0 false false false
		      (create_post_m 0 false false false 
			 (create_post_thread (i+1) sofar)
		      )))))
  )
  in create_post_thread 1 (create_post_thread0 (AC.empty_transitionSet))


let all_locals=
  SetOfComponent.add local_a (SetOfComponent.add local_b (SetOfComponent.add local_c (SetOfComponent.add local_d (SetOfComponent.add local_e (SetOfComponent.add local_f (SetOfComponent.add local_g (SetOfComponent.add local_h (SetOfComponent.add local_i (SetOfComponent.add local_j (SetOfComponent.add local_k (SetOfComponent.add local_l (SetOfComponent.add local_m (SetOfComponent.add local_n (SetOfComponent.empty))))))))))))) )

(* returns the set of all shared states *)
let all_shared n=
  let rec all_shared_aux pio sF sE st sofar =
    if pio>(max_pio_of_thread_number n) then sofar else
      let new_shared=shared_state pio sF sE st in
      let sofar'=SetOfShared.add new_shared sofar in
	try let (sF',sE',st')=three_counter (sF,sE,st) in
	  all_shared_aux pio sF' sE' st' sofar'
	with Done ->
	  all_shared_aux (pio+1) false false false sofar'
  in
    all_shared_aux 0 false false false (SetOfShared.empty)


let st_is_true_shared n =
  let rec st_is_true_shared_aux pio sF sE st sofar =
    if pio>(max_pio_of_thread_number n) then sofar else
      let new_shared=shared_state pio sF sE true in
      let sofar'=SetOfShared.add new_shared sofar in
	try let (sF',sE',st')=three_counter (sF,sE,st) in
	  st_is_true_shared_aux pio sF' sE' st' sofar'
	with Done ->
	  st_is_true_shared_aux (pio+1) false false false sofar'
  in
    st_is_true_shared_aux 0 false false false (SetOfShared.empty)

(* creates error states *)
let create_error n = AC.no_assertion_setOfStates (SetOfComponent.singleton local_j) all_locals (st_is_true_shared n) (n+1)


(* preparing the arugments and starting the model checker. *)
let start_model_checker n=
  let init=AC.singleton_setOfStates (shared_state 1 false false false) (Array.append (Array.make 1 local_a) (Array.make n local_h)) in
  let error=create_error n in
  let post:AC.transitionSet=create_post n in
(*  Debug.print_debug ("Transition set: "^(AC.string_of_transitionSet post)^".\n"); *)
  let mc_result = model_check init error post in
    match mc_result with
	Safe invariant ->print_string ("Safe. Invariant:\n"^(AC.string_of_setOfStates invariant)^".\n")
      | Unsafe trace -> print_string ("Unsafe. Error trace:\n"^(AC.string_of_trace trace)^".\n")


let main () =
  let error_message ()=print_string "Checking the bluetooth driver. Syntax:\nbluetooth <number_of_threads>\n" in
  if ((Array.length(Sys.argv))<>2) then error_message () else
    try 
      let n=int_of_string Sys.argv.(1)
      in start_model_checker n
    with Failure f -> error_message ()
;;
main ()

