(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

type glob = int
type loc = int
type thread_state = glob * loc
type prog_state = glob * (loc array)
type thread_transition = thread_state * int * thread_state

let rec foldl2_array (f: 'a -> ('b*'b) -> 'a) (init:'a) ((bar1:('b array)),(bar2
:('b array))) :'a =
  let
    len = Array.length bar1 and
    len2 = Array.length bar2
  in
    if len <> len2 then raise (Invalid_argument "foldl2_array expects arrays of
equal size") else
    if len=0 then init else
      f (foldl2_array f init (Array.sub bar1 1 (len - 1), Array.sub bar2 1 (len-
1))) (bar1.(0),bar2.(0))

module OrderedProgState =
struct
  type t = prog_state
  let compare ((g1,l1):t) ((g2,l2):t) =
    if g1<g2 then (-1)
    else if g1>g2 then 1
    else foldl2_array
      (fun (cmp:int) ((el1:loc),(el2:loc)) -> if el1>el2 then 1 else if el1<el2 then (-1) else cmp)
      0
      (l1,l2)
end

module OrderedThreadState =
  struct
    type t = thread_state
    let compare ((g1,l1):t) ((g2,l2):t) =
      if g1<g2 then -1 else if g1>g2 then 1 else compare l1 l2 (*if l1<l2 then -1 else if l1>l2 then 1 else 0 *)
  end

module OrderedThreadTransition =
  struct
    type t = thread_transition
    let compare ((g,n,l):thread_transition) ((g',n',l'):thread_transition) :int
=
      if n<n' then -1 else
        if n>n' then 1 else
          if g<g' then -1 else
            if g>g' then 1 else
              compare l l'
              (*if l<l' then -1 else if l>l' then 1 else 0 *)
end

(*
module OrderedLocalState = 
  struct
    type t=loc
    let compare (l:loc) (l':loc) = if l<l' then -1 else if l>l' then 1 else 0
  end

module OrderedGlobalState = 
  struct
    type t=glob
    let compare (g:loc) (g':loc) = if g<g' then -1 else if g>g' then 1 else 0
  end
*)

module PairOfOrderedGlobStr = 
  struct
    type left_type=glob
    type right_type=string
    let compare_left (l:left_type) (l':left_type) = Pervasives.compare l l'
    let compare_right (r:right_type) (r':right_type) = String.compare r r'
  end
  
module PairOfOrderedLocStr = 
  struct
    type left_type=loc
    type right_type=string
    let compare_left (l:left_type) (l':left_type) = Pervasives.compare l l'
    let compare_right (r:right_type) (r':right_type) = String.compare r r'
  end
(*
module PairOfOrderedThreadStr = 
  struct
    type left_type=int
    type right_type=string
    let compare_left (l:left_type) (l':left_type) = Pervasives.compare l l'
    let compare_right (r:right_type) (r':right_type) = String.compare r r'
  end
*)
module SetOfProgStates = Set.Make (OrderedProgState)
type setOfProgStates = SetOfProgStates.t
(*
module SetOfLocals = Set.Make (OrderedLocalState)
type setOfLocals = SetOfLocals.t
module SetOfGlobals = Set.Make (OrderedGlobalState)
type setOfGlobals = SetOfGlobals.t
*)
module SetOfThreadTransitions = Set.Make (OrderedThreadTransition)
type setOfThreadTransitions = SetOfThreadTransitions.t
module BijectiveMapGlobStr = BijectiveMap.Make (PairOfOrderedGlobStr)
type bijectiveMapGlobStr=BijectiveMapGlobStr.t
module BijectiveMapLocStr = BijectiveMap.Make (PairOfOrderedLocStr)
type bijectiveMapLocStr=BijectiveMapLocStr.t
(*
module BijectiveMapThreadStr = BijectiveMap.Make (PairOfOrderedThreadStr)
type bijectiveMapThreadStr=BijectiveMapThreadStr.t
*)
(* Threads werden durchnummeriert von 0 bis no_of_threads-1, globale Zustnde von 0 bis no_of_globals-1, lokale Zustnde von 0 bis no_of_locals-1 *)
type verificationTask = {no_of_globals:int; glob_to_str:bijectiveMapGlobStr; no_of_locals:int; loc_to_str:bijectiveMapLocStr; no_of_threads:int; all_transitions:setOfThreadTransitions; init:setOfProgStates; error:setOfProgStates}


(*
 * String prefix.
 *)
let rec is_string_prefix_aux s1 s2 i len =
   i = len || (String.unsafe_get s1 i = String.unsafe_get s2 i && is_string_prefix_aux s1 s2 (i + 1) len)

let is_string_prefix s1 s2 =
   let len1 = String.length s1 in
   let len2 = String.length s2 in
      len1 <= len2 && is_string_prefix_aux s1 s2 0 len1


(*
 * Split a string str into a list of substrings.
 * The string is split on any character in delims.  Quotations
 * are not split.
 *
 * Empty substrings are _not_ returned as empty strings in the list.
 * For example:
 *   split ".-" "foo.bar--ba??z"
 * returns
 *   ["foo"; "bar"; "ba??z"]
 *)
let tokens_fold f x quotes delims str =
   let strlen = String.length str in
   (* Skip white space *)
   let rec skip_split pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains delims c then skip_split (succ pos)
            else pos
   in
   (* Find the next split index. *)
   let rec next_split pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains delims c then pos
            else if String.contains quotes c then next_quote (succ pos)
            else next_split (succ pos)

   and next_quote pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains quotes c then next_split (succ pos)
            else if c = '\\' && pos < pred strlen then next_quote (pos + 2)
            else next_quote (succ pos)
   in
   (* Build the list *)
   let rec str_split x pos =
      if pos = strlen then x
      else
         let pos_end = next_split pos in
         let x = f x str pos (pos_end - pos) in
            str_split x (skip_split pos_end)
   in str_split x (skip_split 0)

(*
 * Default token processor.
 *)
let tokens quotes delims str =
   let l =
      tokens_fold (fun l s off len ->
            String.sub s off len :: l) [] quotes delims str
   in List.rev l

let tokens_std (s:string): string list = tokens "\"'" " \t\r\n\012" s




(*Reads a verification taks from the standard input*)
let read_verificationTask () : verificationTask = 
  let rec read_verifTask_accum (read_until_now:verificationTask):verificationTask = 
    try
      let input_line = read_line ()
      in let ts = tokens_std input_line
      in let new_verif_task=
	  if ((List.length ts)=0) or ((String.compare "#" (List.hd ts)) = 0) 
	  then read_until_now
	  else 
	    let rec get_local_states list_remainder =
	      (* Convert the list of locals into an array *)
	      if ((List.length list_remainder)=0) then ((Array.make 0 0),(read_until_now.loc_to_str),(read_until_now.no_of_locals)) else 
		let local_str=List.hd list_remainder
		and (remainer_array,new_loc_to_str_map,no_of_locals)=get_local_states (List.tl list_remainder)
		in let (local_state,new_local_state_needed)= 
		    try (BijectiveMapLocStr.find_right local_str new_loc_to_str_map,false)
		    with Not_found  -> ((try ((BijectiveMapLocStr.max_left new_loc_to_str_map)+1) with Not_found -> 0),true)
		in ((Array.append (Array.make 1 local_state) remainer_array),(if new_local_state_needed then BijectiveMapLocStr.add local_state local_str new_loc_to_str_map else new_loc_to_str_map), (if new_local_state_needed then no_of_locals+1 else no_of_locals))
            in if (String.compare "T" (List.hd ts))=0 then 
		(* Input a new transition *)
		(if (List.length ts)<>6 then raise (Invalid_argument ("There is a bad inpit line that starts with T:"^input_line^"\nFormat:\nT <thread_number> <global> <local> <global'> <local'>\n")) else
		   let thread_no_str=List.hd (List.tl ts)
		   and global_str=List.hd (List.tl (List.tl ts))
		   and local_str=List.hd (List.tl (List.tl (List.tl ts)))
		   and global_next_str=List.hd (List.tl (List.tl (List.tl (List.tl ts))))
		   and local_next_str=List.hd (List.tl (List.tl (List.tl (List.tl (List.tl ts)))))
		   in let thread_no=
		       try int_of_string thread_no_str
		       with Failure s -> raise (Invalid_argument ("The input line\n"^input_line^"\nspecifies a non-numeric thread number. Format:\nT <thread_number> <global> <local> <global'> <local'>\n"))
		      and (global_state,new_global_state_needed)=
		       try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		   in let new_glob_to_str=if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
		      and (local_state,new_local_state_needed)= 
		       try (BijectiveMapLocStr.find_right local_str read_until_now.loc_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapLocStr.max_left read_until_now.loc_to_str)+1) with Not_found -> 0),true)
		   in let new_loc_to_str = if new_local_state_needed then BijectiveMapLocStr.add local_state local_str read_until_now.loc_to_str else read_until_now.loc_to_str
		      and (global_next_state,new_global_next_state_needed)=
		       try (BijectiveMapGlobStr.find_right global_next_str new_glob_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapGlobStr.max_left new_glob_to_str)+1) with Not_found -> 0),true)
		   in let new_new_glob_to_str=if new_global_next_state_needed then BijectiveMapGlobStr.add global_next_state global_next_str new_glob_to_str else new_glob_to_str
		      and (local_next_state,new_local_next_state_needed)= 
		       try (BijectiveMapLocStr.find_right local_next_str new_loc_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapLocStr.max_left new_loc_to_str)+1) with Not_found -> 0),true)
		   in let new_new_loc_to_str=if new_local_next_state_needed then BijectiveMapLocStr.add local_next_state local_next_str new_loc_to_str else new_loc_to_str
		      and new_transition=((global_state,local_state),thread_no,(global_next_state,local_next_state))
		   in let new_transition_set=SetOfThreadTransitions.add new_transition read_until_now.all_transitions
		   in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then (if new_global_next_state_needed then 2 else 1) else (if new_global_next_state_needed then 1 else 0)); glob_to_str=new_new_glob_to_str; no_of_locals=read_until_now.no_of_locals +(if new_local_state_needed && new_local_next_state_needed then 2 else (if (new_local_state_needed || new_local_next_state_needed) then 1else 0)); loc_to_str=new_new_loc_to_str; no_of_threads=if read_until_now.no_of_threads<=thread_no then thread_no+1 else  read_until_now.no_of_threads; all_transitions=new_transition_set; init=read_until_now.init; error=read_until_now.error})
	      else if ((String.compare "Init" (List.hd ts))=0) then 
		(* Input a new initial state. *)
		(let number_of_threads=(List.length ts)-2
		 in if number_of_threads<=0 then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nFormat:\nInit <global> <local_1> ... <local_n>\n")) else if number_of_threads<read_until_now.no_of_threads then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nThe number of threads that were read until now is "^(string_of_int read_until_now.no_of_threads)^", which exceeds the number of threads specified by this initial state.\n")) else if ((not (SetOfProgStates.is_empty read_until_now.init)) && ((Array.length (snd (SetOfProgStates.choose read_until_now.init)))<>number_of_threads)) then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nA previously declared initial state had "^(string_of_int (Array.length (snd (SetOfProgStates.choose read_until_now.init)))^" threads.\n"))) else if (not (SetOfProgStates.is_empty read_until_now.error)) && ((Array.length (snd (SetOfProgStates.choose read_until_now.error)))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nA previously declared error state had "^(string_of_int (Array.length (snd (SetOfProgStates.choose read_until_now.error)))^" threads.\n"))) else 
		     let global_str=List.hd (List.tl ts)
		     and locals_list=List.tl (List.tl ts)
		     in let (global_state,new_global_state_needed)=
			 try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
			 with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		     in let new_glob_to_str=if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
			and (locals_array,new_loc_to_str,new_no_of_locals)=get_local_states locals_list
		     in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then 1 else 0); glob_to_str = new_glob_to_str; no_of_locals=new_no_of_locals; loc_to_str=new_loc_to_str; no_of_threads=number_of_threads; all_transitions=read_until_now.all_transitions; init=SetOfProgStates.add (global_state,locals_array) read_until_now.init;error=read_until_now.error})
	      else if (String.compare "Error" (List.hd ts))=0 then 
		(* Input a new error state *)
		(let number_of_threads=(List.length ts)-2
		 in if number_of_threads<=0 then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nFormat:\nError <global> <local_1> ... <local_n>\n")) else if number_of_threads<read_until_now.no_of_threads then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nThe number of threads that were read until now is "^(string_of_int read_until_now.no_of_threads)^", which exceeds the number of threads specified by this initial state.\n")) else if (not (SetOfProgStates.is_empty read_until_now.error)) && ((Array.length (snd (SetOfProgStates.choose read_until_now.error)))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nA previously declared error state had "^(string_of_int (Array.length (snd (SetOfProgStates.choose read_until_now.error)))^" threads.\n"))) else if (not (SetOfProgStates.is_empty read_until_now.init)) && ((Array.length (snd (SetOfProgStates.choose read_until_now.init)))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nA previously declared initial state had "^(string_of_int (Array.length (snd (SetOfProgStates.choose read_until_now.init)))^" threads.\n"))) else
		     let global_str=List.hd (List.tl ts)
		     and locals_list=List.tl (List.tl ts)
		     in let (global_state,new_global_state_needed)=
			 try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
			 with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		     in let new_glob_to_str = if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
			and (locals_array,new_loc_to_str,new_no_of_locals)=get_local_states locals_list
		     in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then 1 else 0); glob_to_str = new_glob_to_str; no_of_locals=new_no_of_locals; loc_to_str=new_loc_to_str; no_of_threads=number_of_threads; all_transitions=read_until_now.all_transitions; init=read_until_now.init; error=SetOfProgStates.add (global_state,locals_array) read_until_now.error})
	      else (raise (Invalid_argument ("In the input, some line starts with an unknown word "^(List.hd ts)^"\n")))
      in read_verifTask_accum new_verif_task
    with End_of_file -> read_until_now
  in read_verifTask_accum {no_of_globals=0; glob_to_str=BijectiveMapGlobStr.empty; no_of_locals=0; loc_to_str=BijectiveMapLocStr.empty; no_of_threads=0; all_transitions=SetOfThreadTransitions.empty; init=SetOfProgStates.empty; error=SetOfProgStates.empty}
       

(* Coverting everything into strings for printing. *)

(* Converts a program state into a string *)
let string_of_prog_state ((g,l):prog_state) (glob_to_str:bijectiveMapGlobStr) (loc_to_str:bijectiveMapLocStr) = 
 ("("^(BijectiveMapGlobStr.find_left g glob_to_str)^","^(String.concat "," (List.map (function ll -> BijectiveMapLocStr.find_left ll loc_to_str) (Array.to_list l)))^")")

(* Converts a set of program states into a string *)
let string_of_setOfProgStates (s:setOfProgStates) (glob_to_str:bijectiveMapGlobStr) (loc_to_str:bijectiveMapLocStr) =
  ("{"^(String.concat ", " (List.map (function ps -> string_of_prog_state ps glob_to_str loc_to_str) (SetOfProgStates.elements s)))^"}")

(* Converts a single thread transition into a string *)
let string_of_thread_transition (((g,l),i,(g',l')):thread_transition) (glob_to_str:bijectiveMapGlobStr) (loc_to_str:bijectiveMapLocStr) =
  ("(("^(BijectiveMapGlobStr.find_left g glob_to_str)^","^(BijectiveMapLocStr.find_left l loc_to_str)^"),"^(string_of_int i)^",("^(BijectiveMapGlobStr.find_left g' glob_to_str)^","^(BijectiveMapLocStr.find_left l' loc_to_str)^"))")


(* Converts a set of thread transitions into string *)
let string_of_setOfThreadTransitions (t: setOfThreadTransitions) (glob_to_str:bijectiveMapGlobStr) (loc_to_str:bijectiveMapLocStr) =
  ("{"^(String.concat ", " (List.map (function tt -> string_of_thread_transition tt glob_to_str loc_to_str) (SetOfThreadTransitions.elements t)))^"}")

(* Converts a verification task into a string *)
let string_of_verificationTask (vt:verificationTask) =
  "Transitions:\n"^(string_of_setOfThreadTransitions vt.all_transitions vt.glob_to_str vt.loc_to_str)^"\nInitial states:\n"^(string_of_setOfProgStates vt.init vt.glob_to_str vt.loc_to_str)^"\nError states:\n"^(string_of_setOfProgStates vt.error vt.glob_to_str vt.loc_to_str)^"\n"



(* Testing input-output: *)
let main () =
  print_string (string_of_verificationTask (read_verificationTask ()))
;;

main ();;
